r"""
Groups of elements representing (complex) arguments.

This includes

- :class:`RootsOfUnityGroup` (containing all roots of unity)

- :class:`UnitCircleGroup` (representing elements on the unit circle by
  `e^{2\pi\cdot\mathit{exponent}}`)

- :class:`ArgumentByElementGroup` (whose elements are defined via
  formal arguments by `e^{I\cdot\mathrm{arg}(\mathit{element})}`.

Use the factory :class:`ArgumentGroup` for creating such a group conveniently.

.. NOTE::

    One main purpose of such groups is in an
    :mod:`asymptotic ring's <sage.rings.asymptotic.asymptotic_ring>`
    :mod:`growth group <sage.rings.asymptotic.growth_group>`
    when an element like `z^n`
    (for some constant `z`) is split into
    `\lvert z \rvert^n \cdot e^{I\cdot \mathrm{arg}(z) n}`.
    (Note that the first factor determines the growth of that product,
    the second does not influence the growth.)

AUTHORS:

- Daniel Krenn (2018)

Classes and Methods
===================
"""
# ****************************************************************************
# Copyright (C) 2018 Daniel Krenn <dev@danielkrenn.at>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import sage.rings.abc
from sage.structure.element import MultiplicativeGroupElement
from sage.structure.factory import UniqueFactory
from sage.structure.parent import Parent
from sage.structure.richcmp import richcmp_by_eq_and_lt
from sage.structure.unique_representation import UniqueRepresentation


class AbstractArgument(MultiplicativeGroupElement):
    r"""
    An element of :class:`AbstractArgumentGroup`. This abstract class
    encapsulates an element of the parent's base, i.e. it can be seen
    as a wrapper class.

    INPUT:

    - ``parent`` -- a SageMath parent

    - ``element`` -- an element of parent's base

    - ``normalize`` -- boolean (default: ``True``)
    """

    def __init__(self, parent, element, normalize=True):
        r"""
        See :class:`AbstractArgument` for more information.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import UnitCircleGroup, RootsOfUnityGroup
            sage: C = UnitCircleGroup(RR)
            sage: C(exponent=1/2)  # indirect doctest
            e^(2*pi*0.500000000000000)
            sage: C(exponent=3/2)
            e^(2*pi*0.500000000000000)

            sage: U = RootsOfUnityGroup()
            sage: U(exponent=0)
            1
            sage: U(exponent=1)
            1
            sage: U(exponent=2/3) == U(exponent=5/3)
            True
        """
        if parent is None:
            raise ValueError('parent must be provided')
        super().__init__(parent=parent)

        try:
            element = parent.base()(element)
        except (TypeError, ValueError) as e:
            from sage.rings.asymptotic.misc import combine_exceptions
            from sage.structure.element import parent as parent_function
            raise combine_exceptions(
                ValueError(
                    '{} ({}) is not in {}'.format(element,
                                                  parent_function(element),
                                                  parent.base())),
                e)

        if normalize:
            element = self._normalize_(element)
        self._element_ = element

    @staticmethod
    def _normalize_(element):
        r"""
        Normalize the given element.

        INPUT:

        - ``element`` -- an element of the parent's base

        OUTPUT: an element

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import AbstractArgument
            sage: AbstractArgument._normalize_(3/2)
            Traceback (most recent call last):
            ...
            NotImplementedError: only implemented in concrete realizations
        """
        raise NotImplementedError('only implemented in concrete realizations')

    def __hash__(self):
        r"""
        Return a hash value of this argument.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import UnitCircleGroup
            sage: C = UnitCircleGroup(RR)
            sage: hash(C(exponent=1/3))  # indirect doctest, random
            42
        """
        return hash((self.parent(), self._element_))

    def _symbolic_(self, R=None):
        r"""
        Return this argument as a symbolic expression.

        INPUT:

        - ``R`` -- (a subring of) the symbolic ring or ``None``.
          The output will be an element of ``R``. If ``None``,
          then the symbolic ring is used.

        OUTPUT: a symbolic expression

        EXAMPLES::

            sage: from sage.groups.misc_gps.argument_groups import AbstractArgument
            sage: class MyArgument(AbstractArgument):
            ....:     @staticmethod
            ....:     def _normalize_(element):
            ....:         return element
            sage: a = MyArgument(ZZ, -1)
            sage: a._symbolic_()
            Traceback (most recent call last):
            ...
            NotImplementedError: only implemented in concrete realizations
        """
        raise NotImplementedError('only implemented in concrete realizations')

    _richcmp_ = richcmp_by_eq_and_lt("_eq_", "_lt_")

    def _eq_(self, other):
        r"""
        Return whether this argument equals ``other``.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import RootsOfUnityGroup
            sage: U = RootsOfUnityGroup()
            sage: U(exponent=0) == U(exponent=1)
            True
            sage: U(exponent=2/3) == U(exponent=5/3)
            True
            sage: U(exponent=2/3) == U(exponent=-2/3)
            False

        ::

            sage: from sage.groups.misc_gps.argument_groups import ArgumentByElementGroup
            sage: C = ArgumentByElementGroup(CC)
            sage: C(I) == C(I)                                                          # needs sage.symbolic
            True

        As we do not have normalization in :class:`ArgumentByElement`,
        then following, although equal, is not equal::

            sage: C(I) == C(2*I)                                                        # needs sage.symbolic
            False
        """
        return self._element_ == other._element_

    def _lt_(self, other):
        r"""
        Raise an error since points on the unit circle are not comparable.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import RootsOfUnityGroup
            sage: U = RootsOfUnityGroup()
            sage: U(exponent=0) < U(exponent=0)  # indirect doctest
            Traceback (most recent call last):
            ...
            RuntimeError: cannot decide '<' for the roots of unity 1 and 1
            sage: U(exponent=0) < U(exponent=1/2)  # indirect doctest
            Traceback (most recent call last):
            ...
            RuntimeError: cannot decide '<' for the roots of unity 1 and -1
            sage: U(exponent=0) > U(exponent=1/2)  # indirect doctest
            Traceback (most recent call last):
            ...
            RuntimeError: cannot decide '<' for the roots of unity -1 and 1
        """
        raise RuntimeError("cannot decide '<' "
                           "for the roots of unity "
                           "{} and {}".format(self, other))

    def _act_on_(self, other, is_left):
        r"""
        Return the action of this argument on ``other``.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import RootsOfUnityGroup
            sage: U = RootsOfUnityGroup()
            sage: U(-1) * 4
            -4
            sage: _.parent()
            Symbolic Ring
            sage: 4 * U(-1)
            -4
            sage: _.parent()
            Symbolic Ring

            sage: P = Permutation([1,2,3])
            sage: U(-1) * P
            Traceback (most recent call last):
            ...
            TypeError: -1 (Group of Roots of Unity) cannot
            (left-)act on [1, 2, 3] (Standard permutations)
            > *previous* TypeError: no canonical coercion from
              Standard permutations to Symbolic Ring

        ::

            sage: from sage.groups.misc_gps.argument_groups import ArgumentByElementGroup
            sage: C = ArgumentByElementGroup(SR)                                        # needs sage.symbolic
            sage: C(-1) * 4                                                             # needs sage.symbolic
            -4
            sage: _.parent()
            Symbolic Ring
            sage: 4 * C(-1)                                                             # needs sage.symbolic
            -4
            sage: _.parent()
            Symbolic Ring
        """
        from sage.symbolic.ring import SR, SymbolicRing

        P = other.parent()
        S = P if isinstance(P, SymbolicRing) else SR
        try:
            other = S.coerce(other)
        except (TypeError, ValueError) as e:
            from sage.rings.asymptotic.misc import combine_exceptions
            raise combine_exceptions(
                TypeError('{} ({}) cannot ({}-)act on '
                          '{} ({})'.format(
                              self, self.parent(),
                              'left' if is_left else 'right',
                              other, P)),
                e)
        return self._symbolic_(S) * other

    def __abs__(self):
        r"""
        Return the absolute value of this argument which equals `1`.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import RootsOfUnityGroup
            sage: U = RootsOfUnityGroup()
            sage: abs(U(exponent=1/4))  # indirect doctest
            1
            sage: _.parent()
            Integer Ring
        """
        from sage.rings.integer_ring import ZZ
        return ZZ.one()


class AbstractArgumentGroup(UniqueRepresentation, Parent):
    r"""
    A group whose elements represent (complex) arguments.

    INPUT:

    - ``base`` -- a SageMath parent

    - ``category`` -- a category
    """

    Element = AbstractArgument

    @staticmethod
    def __classcall__(cls, base, category=None):
        r"""
        See :class:`AbstractArgumentGroup` for more information.

        TESTS:

            sage: from sage.groups.misc_gps.argument_groups import UnitCircleGroup
            sage: UnitCircleGroup(RR).category()  # indirect doctest
            Category of commutative groups
        """
        category = cls._determine_category_(category)
        return super().__classcall__(cls, base, category)

    @staticmethod
    def _determine_category_(category):
        r"""
        Return the category of this argument group.

        INPUT:

        - ``category`` -- a category or ``None`` (in which case the output
          equals ``category``)

        OUTPUT: a category

        EXAMPLES::

            sage: from sage.groups.misc_gps.argument_groups import UnitCircleGroup
            sage: UnitCircleGroup._determine_category_(None)
            Category of commutative groups
            sage: UnitCircleGroup._determine_category_(Groups())
            Category of groups
        """
        if category is None:
            from sage.categories.groups import Groups
            category = Groups().Commutative()
        return category

    def __init__(self, base, category):
        r"""
        See :class:`AbstractArgumentGroup` for more information.

        TESTS:

            sage: from sage.groups.misc_gps.argument_groups import UnitCircleGroup
            sage: UnitCircleGroup(RR).base()  # indirect doctest
            Real Field with 53 bits of precision
        """
        super().__init__(category=category, base=base)

    def __hash__(self):
        r"""
        Return a hash value of this argument group.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import UnitCircleGroup
            sage: hash(UnitCircleGroup(RR))  # indirect doctest, random
            42
        """
        return hash((self.__class__, self.base()))

    def _an_element_(self):
        r"""
        Return an element of this argument group.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import UnitCircleGroup
            sage: UnitCircleGroup(RR).an_element()  # indirect doctest
            e^(2*pi*0.000000000000000)
        """
        return self.element_class(self, self.base().an_element())


class UnitCirclePoint(AbstractArgument):
    r"""
    An element of :class:`UnitCircleGroup`
    which is `e^{2\pi\cdot\mathit{exponent}}`.

    INPUT:

    - ``parent`` -- a SageMath parent

    - ``exponent`` -- a number (of a subset of the reals)

    - ``normalize`` -- boolean (default: ``True``)
    """

    @staticmethod
    def _normalize_(exponent):
        r"""
        Normalize the given exponent so that it is in `[0,1)`.

        INPUT:

        - ``exponent`` -- an element of the parent's base

        OUTPUT: an element

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import UnitCirclePoint
            sage: UnitCirclePoint._normalize_(3/2)
            1/2
        """
        return exponent - exponent.floor()

    @property
    def exponent(self):
        r"""
        The exponent of this point on the unit circle.

        EXAMPLES::

            sage: from sage.groups.misc_gps.argument_groups import UnitCircleGroup
            sage: C = UnitCircleGroup(RR)
            sage: C(exponent=4/3).exponent
            0.333333333333333
        """
        return self._element_

    def _repr_(self):
        r"""
        Return a representation string of this point on the unit circle.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import UnitCircleGroup
            sage: C = UnitCircleGroup(RR)
            sage: C(exponent=1/3)
            e^(2*pi*0.333333333333333)
        """
        return 'e^(2*pi*{})'.format(self.exponent)

    def _symbolic_(self, R=None):
        r"""
        Return this point on the unit circle as a symbolic expression.

        INPUT:

        - ``R`` -- (a subring of) the symbolic ring or ``None``.
          The output will be an element of ``R``. If ``None``,
          then the symbolic ring is used.

        OUTPUT: a symbolic expression

        EXAMPLES::

            sage: from sage.groups.misc_gps.argument_groups import UnitCircleGroup
            sage: C = UnitCircleGroup(RR)
            sage: C(exponent=1/4)._symbolic_()
            e^(0.500000000000000*I*pi)
            sage: _.parent()
            Symbolic Ring

            sage: from sage.groups.misc_gps.argument_groups import RootsOfUnityGroup
            sage: U = RootsOfUnityGroup()
            sage: U(exponent=1/4)._symbolic_()
            I
            sage: _.parent()
            Symbolic Ring
        """
        from sage.functions.log import exp
        from sage.symbolic.ring import SR

        if R is None:
            R = SR

        return exp(2 * R('pi') * R('I') * self.exponent)

    def _mul_(self, other):
        r"""
        Return the product of this point on the unit circle and ``other``.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import UnitCircleGroup
            sage: C = UnitCircleGroup(RR)
            sage: C(exponent=0.3) * C(exponent=0.4)
            e^(2*pi*0.700000000000000)
        """
        P = self.parent()
        return P.element_class(P, self.exponent + other.exponent)

    def __pow__(self, exponent):
        r"""
        Return the power of this point on the unit circle
        to the given ``exponent``.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import UnitCircleGroup, RootsOfUnityGroup

            sage: C = UnitCircleGroup(RR)
            sage: C(exponent=0.1)^2
            e^(2*pi*0.200000000000000)
            sage: _.parent()
            Unit Circle Group with Exponents in
            Real Field with 53 bits of precision modulo ZZ
            sage: C(exponent=0.1)^QQ(2/1)
            e^(2*pi*0.200000000000000)
            sage: _.parent()
            Unit Circle Group with Exponents in
            Real Field with 53 bits of precision modulo ZZ

            sage: U = RootsOfUnityGroup()
            sage: a = U(exponent=1/7); a
            zeta7
            sage: a^(7/3)
            zeta3
            sage: _.parent()
            Group of Roots of Unity

            sage: U(exponent=1/3)^(0.25)
            e^(2*pi*0.0833333333333333)
            sage: _.parent()
            Unit Circle Group with Exponents in
            Real Field with 53 bits of precision modulo ZZ

            sage: U(exponent=1/3)^x
            (1/2*I*sqrt(3) - 1/2)^x
            sage: U(exponent=1/2)^x
            (-1)^x
            sage: U(exponent=1/4)^x
            I^x
            sage: U(exponent=1/4)^SR(8)
            1
        """
        from sage.symbolic.ring import SymbolicRing

        new_exponent = self.exponent * exponent
        parent = new_exponent.parent()
        if isinstance(new_exponent.parent(), SymbolicRing):
            return self._symbolic_(parent) ** exponent
        return self.parent()._create_element_in_extension_(new_exponent)

    def __invert__(self):
        r"""
        Return the inverse of this point on the unit circle.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import UnitCircleGroup
            sage: C = UnitCircleGroup(RR)
            sage: ~C(exponent=0.4)
            e^(2*pi*0.600000000000000)
            sage: C(1) / C(exponent=0.4)
            e^(2*pi*0.600000000000000)
            sage: C(exponent=0) / C(exponent=0.42)
            e^(2*pi*0.580000000000000)
        """
        P = self.parent()
        return P.element_class(P, -self.exponent)

    def is_one(self):
        r"""
        Return whether this point on the unit circle is `1`.

        EXAMPLES::

            sage: from sage.groups.misc_gps.argument_groups import UnitCircleGroup
            sage: C = UnitCircleGroup(QQ)
            sage: C(exponent=0).is_one()
            True
            sage: C(exponent=1/2).is_one()
            False
            sage: C(exponent=2/3).is_one()
            False
            sage: C(exponent=42).is_one()
            True
        """
        return self.exponent == 0

    def is_minus_one(self):
        r"""
        Return whether this point on the unit circle is `-1`.

        EXAMPLES::

            sage: from sage.groups.misc_gps.argument_groups import UnitCircleGroup
            sage: C = UnitCircleGroup(QQ)
            sage: C(exponent=0).is_minus_one()
            False
            sage: C(exponent=1/2).is_minus_one()
            True
            sage: C(exponent=2/3).is_minus_one()
            False
        """
        from sage.rings.rational_field import QQ
        return self.exponent == QQ((1, 2))


class UnitCircleGroup(AbstractArgumentGroup):
    r"""
    A group of points on the unit circle. These points are
    represented by `e^{2\pi\cdot\mathit{exponent}}`.

    INPUT:

    - ``base`` -- a SageMath parent representing a subset of the reals

    - ``category`` -- a category

    EXAMPLES::

        sage: from sage.groups.misc_gps.argument_groups import UnitCircleGroup

        sage: R = UnitCircleGroup(RR); R
        Unit Circle Group with Exponents in Real Field with 53 bits of precision modulo ZZ
        sage: R(exponent=2.42)
        e^(2*pi*0.420000000000000)

        sage: Q = UnitCircleGroup(QQ); Q
        Unit Circle Group with Exponents in Rational Field modulo ZZ
        sage: Q(exponent=6/5)
        e^(2*pi*1/5)
    """

    Element = UnitCirclePoint

    def _repr_(self):
        r"""
        Return a representation string of this unit circle group.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import UnitCircleGroup
            sage: UnitCircleGroup(RR)  # indirect doctest
            Unit Circle Group with Exponents in Real Field with 53 bits of precision modulo ZZ
        """
        return 'Unit Circle Group with Exponents in {} modulo ZZ'.format(self.base())

    def _repr_short_(self):
        r"""
        Return a short representation string of this unit circle group.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import UnitCircleGroup
            sage: UnitCircleGroup(RR)._repr_short_()
            'UU_RR'
        """
        from sage.rings.asymptotic.misc import parent_to_repr_short
        s = parent_to_repr_short(self.base())
        if ' ' in s:
            s = '({})'.format(s)
        return 'UU_{}'.format(s)

    def _element_constructor_(self, data, exponent=None, **kwds):
        r"""
        Construct an element out of the given data.

        INPUT:

        - ``data`` -- an object

        - ``exponent`` -- a number (of a subset of the reals) or ``None``

        - ``kwds`` -- are passed on to element

        OUTPUT: a :class:`UnitCirclePoint`

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import UnitCircleGroup, RootsOfUnityGroup, SignGroup
            sage: R = UnitCircleGroup(RR)
            sage: R(exponent=1/2)
            e^(2*pi*0.500000000000000)

            sage: U = RootsOfUnityGroup()
            sage: U(exponent=0)
            1
            sage: U(exponent=1)
            1
            sage: U(exponent=1/2)
            -1
            sage: U(exponent=1/4)
            I
            sage: U(exponent=1/3)
            zeta3

            sage: C.<z> = CyclotomicField(6)                                            # needs sage.rings.number_field
            sage: z, U(z)                                                               # needs sage.rings.number_field
            (z, zeta6)
            sage: z^2, U(z^2)                                                           # needs sage.rings.number_field
            (z - 1, zeta3)

            sage: U(ZZ(-1))
            -1
            sage: U(QQ(-1))
            -1
            sage: U(int(-1))
            -1

            sage: S = SignGroup()
            sage: R(S(1))
            e^(2*pi*0.000000000000000)
            sage: R(S(-1))
            e^(2*pi*0.500000000000000)
            sage: U(S(1))
            1
            sage: U(S(-1))
            -1

            sage: R(U(exponent=1/3))
            e^(2*pi*0.333333333333333)

            sage: U(exponent=5/2, normalize=False)
            zeta2^5
        """
        import sage.rings.abc
        from sage.groups.generic import discrete_log
        from sage.rings.asymptotic.misc import combine_exceptions
        from sage.rings.rational_field import QQ

        if exponent is None:
            if isinstance(data, int) and data == 0:
                raise ValueError('no input specified')

            elif isinstance(data, self.element_class):
                if data.parent() == self:
                    return data
                exponent = data.exponent

            elif data == 1 or data == '1':
                exponent = 0

            elif data == -1 or data == '-1':
                exponent = QQ((1, 2))

            else:
                try:
                    P = data.parent()
                except AttributeError:
                    raise TypeError('{} is not in {}'.format(data, self))

                if isinstance(P, SignGroup):
                    if data.is_one():
                        exponent = 0
                    elif data.is_minus_one():
                        exponent = QQ((1, 2))

                elif isinstance(P, UnitCircleGroup):
                    exponent = data.exponent

                elif isinstance(P, sage.rings.abc.NumberField_cyclotomic):
                    zeta = P.gen()
                    n = zeta.multiplicative_order()
                    try:
                        exponent = QQ(discrete_log(data, zeta)) / QQ(n)
                    except ValueError as e:
                        raise combine_exceptions(
                            ValueError('{} is not in {}'.format(data, self)), e)

            if exponent is None:
                raise ValueError('{} is not in {}'.format(data, self))

        elif not isinstance(data, int) or data != 0:
            raise ValueError('input is ambiguous: '
                             '{} as well as exponent={} '
                             'specified'.format(data, exponent))

        return self.element_class(self, exponent, **kwds)

    def _create_element_in_extension_(self, exponent):
        r"""
        Create an element in an extension of this unit circle group which
        is chosen according to the input ``exponent``.

        INPUT:

        - ``exponent`` -- the element data

        OUTPUT: an element

        EXAMPLES::

            sage: from sage.groups.misc_gps.argument_groups import UnitCircleGroup, RootsOfUnityGroup

            sage: C = UnitCircleGroup(QQ)
            sage: C._create_element_in_extension_(2.12).parent()                        # needs sage.rings.number_field
            Unit Circle Group with Exponents in
            Real Field with 53 bits of precision modulo ZZ

            sage: U = RootsOfUnityGroup()
            sage: U._create_element_in_extension_(2.12).parent()                        # needs sage.rings.number_field
            Unit Circle Group with Exponents in
            Real Field with 53 bits of precision modulo ZZ
        """
        if exponent.parent() is self.base():
            parent = self
        else:
            parent = ArgumentGroup(exponents=exponent.parent())
        return parent(exponent=exponent)

    def _coerce_map_from_(self, R):
        r"""
        Return whether ``R`` coerces into this unit circle group.

        INPUT:

        - ``R`` -- a parent

        OUTPUT: boolean

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import UnitCircleGroup, RootsOfUnityGroup, SignGroup
            sage: R = UnitCircleGroup(RR)
            sage: Q = UnitCircleGroup(QQ)
            sage: U = RootsOfUnityGroup()
            sage: S = SignGroup()
            sage: for A in (S, U, Q, R):  # indirect doctest
            ....:     for B in (S, U, Q, R):
            ....:         print('{} has {}coerce map from {}'.format(
            ....:             A,
            ....:             '' if A.has_coerce_map_from(B) else 'no ',
            ....:             B))
            Sign Group
              has coerce map from
              Sign Group
            Sign Group
              has no coerce map from
              Group of Roots of Unity
            Sign Group
              has no coerce map from
              Unit Circle Group with Exponents in Rational Field modulo ZZ
            Sign Group
              has no coerce map from
              Unit Circle Group with Exponents in Real Field with 53 bits of precision modulo ZZ
            Group of Roots of Unity
              has coerce map from
              Sign Group
            Group of Roots of Unity
              has coerce map from
              Group of Roots of Unity
            Group of Roots of Unity
              has coerce map from
              Unit Circle Group with Exponents in Rational Field modulo ZZ
            Group of Roots of Unity
              has no coerce map from
              Unit Circle Group with Exponents in Real Field with 53 bits of precision modulo ZZ
            Unit Circle Group with Exponents in Rational Field modulo ZZ
              has coerce map from
              Sign Group
            Unit Circle Group with Exponents in Rational Field modulo ZZ
              has coerce map from
              Group of Roots of Unity
            Unit Circle Group with Exponents in Rational Field modulo ZZ
              has coerce map from
              Unit Circle Group with Exponents in Rational Field modulo ZZ
            Unit Circle Group with Exponents in Rational Field modulo ZZ
              has no coerce map from
              Unit Circle Group with Exponents in Real Field with 53 bits of precision modulo ZZ
            Unit Circle Group with Exponents in Real Field with 53 bits of precision modulo ZZ
              has coerce map from
              Sign Group
            Unit Circle Group with Exponents in Real Field with 53 bits of precision modulo ZZ
              has coerce map from
              Group of Roots of Unity
            Unit Circle Group with Exponents in Real Field with 53 bits of precision modulo ZZ
              has coerce map from
              Unit Circle Group with Exponents in Rational Field modulo ZZ
            Unit Circle Group with Exponents in Real Field with 53 bits of precision modulo ZZ
              has coerce map from
              Unit Circle Group with Exponents in Real Field with 53 bits of precision modulo ZZ
        """
        if isinstance(R, UnitCircleGroup):
            return self.base().has_coerce_map_from(R.base())
        if isinstance(R, SignGroup):
            return True


class RootOfUnity(UnitCirclePoint):
    r"""
    A root of unity (i.e. an element of :class:`RootsOfUnityGroup`)
    which is `e^{2\pi\cdot\mathit{exponent}}` for a rational ``exponent``.
    """

    def exponent_numerator(self):
        r"""
        Return the numerator of the rational quotient in `[0,1)`
        representing the exponent of this root of unity.

        EXAMPLES::

            sage: from sage.groups.misc_gps.argument_groups import RootsOfUnityGroup
            sage: U = RootsOfUnityGroup()
            sage: a = U(exponent=2/3); a
            zeta3^2
            sage: a.exponent_numerator()
            2
        """
        return self.exponent.numerator()

    def exponent_denominator(self):
        r"""
        Return the denominator of the rational quotient in `[0,1)`
        representing the exponent of this root of unity.

        EXAMPLES::

            sage: from sage.groups.misc_gps.argument_groups import RootsOfUnityGroup
            sage: U = RootsOfUnityGroup()
            sage: a = U(exponent=2/3); a
            zeta3^2
            sage: a.exponent_denominator()
            3
        """
        return self.exponent.denominator()

    def _repr_(self):
        r"""
        Return a representation string of this root of unity.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import RootsOfUnityGroup
            sage: U = RootsOfUnityGroup()
            sage: U(exponent=0)
            1
            sage: U(exponent=1/2)
            -1
            sage: U(exponent=1/4)
            I
            sage: U(exponent=3/4)
            -I
            sage: U(exponent=1/3)
            zeta3
            sage: U(exponent=2/3)
            zeta3^2
        """
        from sage.rings.rational_field import QQ
        if self.exponent == 0:
            return '1'
        if self.exponent == QQ((1, 2)):
            return '-1'
        if self.exponent == QQ((1, 4)):
            return 'I'
        if self.exponent == QQ((3, 4)):
            return '-I'
        num = self.exponent_numerator()
        den = self.exponent_denominator()
        zeta = 'zeta{}'.format(den)
        if num == 1:
            return zeta
        return '{}^{}'.format(zeta, num)


class RootsOfUnityGroup(UnitCircleGroup):
    r"""
    The group of all roots of unity.

    INPUT:

    - ``category`` -- a category

    This is a specialized :class:`UnitCircleGroup` with base `\QQ`.

    EXAMPLES::

        sage: from sage.groups.misc_gps.argument_groups import RootsOfUnityGroup
        sage: U = RootsOfUnityGroup(); U
        Group of Roots of Unity
        sage: U(exponent=1/4)
        I
    """

    Element = RootOfUnity

    @staticmethod
    def __classcall__(cls, category=None):
        r"""
        See :class:`RootsOfUnityGroup` for more information.

        TESTS:

            sage: from sage.groups.misc_gps.argument_groups import RootsOfUnityGroup
            sage: RootsOfUnityGroup().category()  # indirect doctest
            Category of commutative groups
        """
        category = cls._determine_category_(category)
        return super(AbstractArgumentGroup, cls).__classcall__(cls, category)

    def __init__(self, category):
        r"""
        See :class:`RootsOfUnityGroup` for more information.

        TESTS:

            sage: from sage.groups.misc_gps.argument_groups import RootsOfUnityGroup
            sage: RootsOfUnityGroup().base()  # indirect doctest
            Rational Field
        """
        from sage.rings.rational_field import QQ
        super().__init__(base=QQ, category=category)

    def _repr_(self):
        r"""
        Return a representation string of this roots of unity group.

        TESTS:

            sage: from sage.groups.misc_gps.argument_groups import RootsOfUnityGroup
            sage: RootsOfUnityGroup()  # indirect doctest
            Group of Roots of Unity
        """
        return 'Group of Roots of Unity'

    def _repr_short_(self):
        r"""
        Return a short representation string of this roots of unity group.

        TESTS:

            sage: from sage.groups.misc_gps.argument_groups import RootsOfUnityGroup
            sage: RootsOfUnityGroup()._repr_short_()
            'UU'
        """
        return 'UU'


class ArgumentByElement(AbstractArgument):
    r"""
    An element of :class:`ArgumentByElementGroup`.

    INPUT:

    - ``parent`` -- a SageMath parent

    - ``element`` -- a nonzero element of the parent's base

    - ``normalize`` -- boolean (default: ``True``)
    """

    def __init__(self, parent, element, normalize=True):
        r"""
        See :class:`ArgumentByElement` for more information.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import ArgumentByElementGroup
            sage: C = ArgumentByElementGroup(CC)
            sage: C(1+2*I)  # indirect doctest                                          # needs sage.symbolic
            e^(I*arg(1.00000000000000 + 2.00000000000000*I))
        """
        super().__init__(parent, element, normalize=normalize)
        if self._element_ == 0:
            raise ValueError('{} is not allowed'.format(element))

    @staticmethod
    def _normalize_(element):
        r"""
        Normalize the given element.

        This is the identity for :class:`ArgumentByElement`.

        INPUT:

        - ``element`` -- an element of the parent's base

        OUTPUT: an element

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import ArgumentByElement
            sage: ArgumentByElement._normalize_(3/2)
            3/2
        """
        return element

    def _repr_(self):
        r"""
        Return a representation string of this argument by element.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import ArgumentByElementGroup
            sage: C = ArgumentByElementGroup(CC)
            sage: C(2+3*I)  # indirect doctest                                          # needs sage.symbolic
            e^(I*arg(2.00000000000000 + 3.00000000000000*I))
        """
        return 'e^(I*arg({}))'.format(self._element_)

    def _symbolic_(self, R=None):
        r"""
        Return this argument by element as a symbolic expression.

        INPUT:

        - ``R`` -- (a subring of) the symbolic ring or ``None``.
          The output will be an element of ``R``. If ``None``,
          then the symbolic ring is used.

        OUTPUT: a symbolic expression

        EXAMPLES::

            sage: from sage.groups.misc_gps.argument_groups import ArgumentByElementGroup
            sage: C = ArgumentByElementGroup(ZZ)
            sage: C(-2)._symbolic_()                                                    # needs sage.symbolic
            -1
            sage: _.parent()
            Symbolic Ring
        """
        from sage.functions.log import exp
        from sage.functions.other import arg
        from sage.symbolic.ring import SR

        if R is None:
            R = SR

        return exp(R('I') * arg(self._element_))

    def _mul_(self, other):
        r"""
        Return the product of this argument by element with ``other``.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import ArgumentByElementGroup
            sage: C = ArgumentByElementGroup(CC)
            sage: C(I) * C(1 + I)  # indirect doctest                                   # needs sage.symbolic
            e^(I*arg(-1.00000000000000 + 1.00000000000000*I))
        """
        P = self.parent()
        return P.element_class(P, self._element_ * other._element_)

    def __pow__(self, exponent):
        r"""
        Return the power of this argument by element
        to the given ``exponent``.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import ArgumentByElementGroup
            sage: C = ArgumentByElementGroup(CC)
            sage: C(I)^5  # indirect doctest                                            # needs sage.symbolic
            e^(I*arg(1.00000000000000*I))
            sage: _.parent()
            Unit Circle Group with Argument of Elements in
            Complex Field with 53 bits of precision
            sage: C(1+I)^3  # indirect doctest                                          # needs sage.symbolic
            e^(I*arg(-2.00000000000000 + 2.00000000000000*I))
            sage: _.parent()
            Unit Circle Group with Argument of Elements in
            Complex Field with 53 bits of precision

            sage: C = ArgumentByElementGroup(RR)
            sage: C(0.42)^CC(2.4)
            e^(I*arg(0.124680431591996))
            sage: _.parent()
            Unit Circle Group with Argument of Elements in
            Complex Field with 53 bits of precision

            sage: C = ArgumentByElementGroup(QQ)
            sage: a = C(-20)^x; a
            (-1)^x
            sage: a.parent()
            Symbolic Ring
        """
        from sage.symbolic.ring import SymbolicRing

        element = self._element_ ** exponent
        parent = element.parent()
        if isinstance(parent, SymbolicRing):
            return self._symbolic_(parent) ** exponent
        return self.parent()._create_element_in_extension_(element)

    def __invert__(self):
        r"""
        Return the inverse of this argument by element.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import ArgumentByElementGroup
            sage: C = ArgumentByElementGroup(CC)
            sage: ~C(I)  # indirect doctest                                             # needs sage.symbolic
            e^(I*arg(-1.00000000000000*I))
        """
        P = self.parent()
        return P.element_class(P, ~self._element_)


class ArgumentByElementGroup(AbstractArgumentGroup):
    r"""
    A group of (complex) arguments. The arguments are represented
    by a the formal argument of an element, i.e.,
    by `\mathrm{arg}(\mathit{element})`.

    INPUT:

    - ``base`` -- a SageMath parent representing a subset of the complex plane

    - ``category`` -- a category

    EXAMPLES::

        sage: from sage.groups.misc_gps.argument_groups import ArgumentByElementGroup
        sage: C = ArgumentByElementGroup(CC); C
        Unit Circle Group with Argument of Elements in
        Complex Field with 53 bits of precision
        sage: C(1 + 2*I)                                                                # needs sage.symbolic
        e^(I*arg(1.00000000000000 + 2.00000000000000*I))
    """

    Element = ArgumentByElement

    def _repr_(self):
        r"""
        Return a representation string of this argument by element group.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import ArgumentByElementGroup
            sage: ArgumentByElementGroup(CC)  # indirect doctest
            Unit Circle Group with Argument of Elements in
            Complex Field with 53 bits of precision
        """
        return 'Unit Circle Group with Argument of Elements in {}'.format(self.base())

    def _repr_short_(self):
        r"""
        Return a short representation string of this  argument by element group.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import ArgumentByElementGroup
            sage: ArgumentByElementGroup(CC)._repr_short_()
            'Arg_CC'
        """
        from sage.rings.asymptotic.misc import parent_to_repr_short, repr_op
        return repr_op('Arg', '_', parent_to_repr_short(self.base()))

    def _element_constructor_(self, data, **kwds):
        r"""
        Construct an element out of the given data.

        INPUT:

        - ``data`` -- an object

        - ``kwds`` -- are passed on to element

        OUTPUT: a :class:`ArgumentByElement`

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import ArgumentByElementGroup
            sage: C = ArgumentByElementGroup(CC)
            sage: C(1 + 2*I)  # indirect doctest                                        # needs sage.symbolic
            e^(I*arg(1.00000000000000 + 2.00000000000000*I))
            sage: C(1)
            e^(I*arg(1.00000000000000))
            sage: C(-1)
            e^(I*arg(-1.00000000000000))
            sage: C(ZZ(-1))
            e^(I*arg(-1.00000000000000))
            sage: C(QQ(-1))
            e^(I*arg(-1.00000000000000))
            sage: C(int(-1))
            e^(I*arg(-1.00000000000000))
            sage: C('-1')
            e^(I*arg(-1.00000000000000))

            sage: from sage.groups.misc_gps.argument_groups import SignGroup
            sage: S = SignGroup()
            sage: C(S(1))
            e^(I*arg(1.00000000000000))
            sage: C(S(-1))
            e^(I*arg(-1.00000000000000))
        """
        from sage.structure.element import parent

        if isinstance(data, int) and data == 0:
            raise ValueError('no input specified')

        elif isinstance(data, self.element_class):
            if data.parent() == self:
                return data
            element = data._element_

        elif data == '1':
            element = 1

        elif data == '-1':
            element = -1

        else:
            P = parent(data)

            if isinstance(P, SignGroup):
                if data.is_one():
                    element = 1
                elif data.is_minus_one():
                    element = -1

            else:
                element = data

        return self.element_class(self, element, **kwds)

    def _create_element_in_extension_(self, element):
        r"""
        Create an element in an extension of this
        argument by element group which
        is chosen according to the input ``element``.

        INPUT:

        - ``element`` -- the element data

        OUTPUT: an element

        EXAMPLES::

            sage: from sage.groups.misc_gps.argument_groups import ArgumentByElementGroup
            sage: C = ArgumentByElementGroup(ZZ)
            sage: C._create_element_in_extension_(2/3).parent()
            Unit Circle Group with Argument of Elements in
            Rational Field
            sage: C._create_element_in_extension_(0.23).parent()
            Unit Circle Group with Argument of Elements in
            Real Field with 53 bits of precision

            sage: C = ArgumentByElementGroup(RR)
            sage: C._create_element_in_extension_(CC(0.23)).parent()
            Unit Circle Group with Argument of Elements in
            Complex Field with 53 bits of precision
        """
        if element.parent() is self.base():
            parent = self
        else:
            parent = ArgumentByElementGroup(element.parent())
        return parent(element)

    def _coerce_map_from_(self, R):
        r"""
        Return whether ``R`` coerces into this argument by element group.

        INPUT:

        - ``R`` -- a parent

        OUTPUT: boolean

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import ArgumentByElementGroup
            sage: from sage.groups.misc_gps.argument_groups import SignGroup
            sage: C = ArgumentByElementGroup(ZZ)
            sage: S = SignGroup()
            sage: C.has_coerce_map_from(S)  # indirect doctest
            True
        """
        if isinstance(R, SignGroup):
            return True


class Sign(AbstractArgument):
    r"""
    An element of :class:`SignGroup`.

    INPUT:

    - ``parent`` -- a SageMath parent

    - ``element`` -- a nonzero element of the parent's base

    - ``normalize`` -- boolean (default: ``True``)
    """

    def __init__(self, parent, element, normalize=True):
        r"""
        See :class:`Sign` for more information.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import SignGroup
            sage: S = SignGroup()
            sage: S.an_element()  # indirect doctest
            -1
        """
        super().__init__(parent, int(element), normalize=normalize)
        if self._element_ not in (-1, 1):
            raise ValueError('{} is not allowed '
                             '(only -1 or 1 is)'.format(element))

    @staticmethod
    def _normalize_(element):
        r"""
        Normalize the given element.

        This is the identity for :class:`Sign`.

        INPUT:

        - ``element`` -- an element of the parent's base

        OUTPUT: an element

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import Sign
            sage: Sign._normalize_(True)
            True
        """
        return element

    def _repr_(self):
        r"""
        Return a representation string of this sign.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import SignGroup
            sage: S = SignGroup()
            sage: S(1)
            1
            sage: S(-1)
            -1
        """
        return repr(self._element_)

    def _mul_(self, other):
        r"""
        Return the product of this sign with ``other``.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import SignGroup
            sage: S = SignGroup()
            sage: S(1) * S(-1)  # indirect doctest
            -1
        """
        P = self.parent()
        return P.element_class(P, self._element_ * other._element_)

    def __pow__(self, exponent):
        r"""
        Return the power of this sign to the given ``exponent``.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import SignGroup
            sage: S = SignGroup()
            sage: S(-1)^4  # indirect doctest
            1
            sage: S(-1)^3  # indirect doctest
            -1

        Check that the results may live in other parents too::

            sage: x = SR.var('x')
            sage: elem = S(-1)^x; elem  # indirect doctest
            (-1)^x
            sage: elem.parent()
            Symbolic Ring

        """
        result = self._element_ ** exponent
        P = self.parent()
        try:
            return P.element_class(P, result)
        except (ValueError, TypeError):
            return result

    def __invert__(self):
        r"""
        Return the inverse of this sign.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import SignGroup
            sage: S = SignGroup()
            sage: ~S(-1)  # indirect doctest
            -1
            sage: _.parent()
            Sign Group
        """
        return self

    def _act_on_(self, other, is_left):
        r"""
        Return the action of this sign on ``other``.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import SignGroup
            sage: S = SignGroup()
            sage: S(-1) * 4
            -4
            sage: _.parent()
            Integer Ring
            sage: 4 * S(-1)
            -4
            sage: _.parent()
            Integer Ring

            sage: S(-1) * ZZ(4)
            -4
            sage: _.parent()
            Integer Ring
            sage: S(-1) * int(4)
            -4
            sage: type(_)
            <class 'int'>
            sage: S(-1) * QQ(4)
            -4
            sage: _.parent()
            Rational Field
            sage: S(-1) * RR(4)
            -4.00000000000000
            sage: _.parent()
            Real Field with 53 bits of precision
            sage: S(-1) * CC(4)
            -4.00000000000000
            sage: _.parent()
            Complex Field with 53 bits of precision
            sage: S(-1) * SR.var('x')                                                   # needs sage.symbolic
            -x
            sage: _.parent()
            Symbolic Ring

        ::

            sage: P = Permutation([1,2,3])
            sage: S(-1) * P
            Traceback (most recent call last):
            ...
            TypeError: unsupported operand parent
            for unary -: 'Standard permutations'
        """
        if self.is_one():
            return other
        if self.is_minus_one():
            return -other

    def is_one(self):
        r"""
        Return whether this sign is `1`.

        EXAMPLES::

            sage: from sage.groups.misc_gps.argument_groups import SignGroup
            sage: S = SignGroup()
            sage: S(-1).is_one()
            False
            sage: S(1).is_one()
            True
        """
        return self._element_ == 1

    def is_minus_one(self):
        r"""
        Return whether this sign is `-1`.

        EXAMPLES::

            sage: from sage.groups.misc_gps.argument_groups import SignGroup
            sage: S = SignGroup()
            sage: S(1).is_minus_one()
            False
            sage: S(-1).is_minus_one()
            True
        """
        return self._element_ == -1


class SignGroup(AbstractArgumentGroup):
    r"""
    A group of the signs `-1` and `1`.

    INPUT:

    - ``category`` -- a category

    EXAMPLES::

        sage: from sage.groups.misc_gps.argument_groups import SignGroup
        sage: S = SignGroup(); S
        Sign Group
        sage: S(-1)
        -1
    """

    Element = Sign

    @staticmethod
    def __classcall__(cls, category=None):
        r"""
        See :class:`SignGroup` for more information.

        TESTS:

            sage: from sage.groups.misc_gps.argument_groups import SignGroup
            sage: S = SignGroup()
            sage: S.category()  # indirect doctest
            Category of finite commutative groups
        """
        category = cls._determine_category_(category).Finite()
        return super(AbstractArgumentGroup, cls).__classcall__(
            cls, category)

    def __init__(self, category):
        r"""
        See :class:`SignGroup` for more information.

        TESTS:

            sage: from sage.groups.misc_gps.argument_groups import SignGroup
            sage: S = SignGroup()
            sage: S.base()  # indirect doctest
            <class 'int'>
        """
        super().__init__(base=int, category=category)

    def _repr_(self):
        r"""
        Return a representation string of the sign group.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import SignGroup
            sage: SignGroup()  # indirect doctest
            Sign Group
        """
        return 'Sign Group'

    def _repr_short_(self):
        r"""
        Return a short representation string of this sign group.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import SignGroup
            sage: S = SignGroup()
            sage: S._repr_short_()
            'Signs'
        """
        return 'Signs'

    def _an_element_(self):
        r"""
        Return a short representation string of this sign group.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import SignGroup
            sage: S = SignGroup()
            sage: S.an_element()
            -1
        """
        return self.element_class(self, -1)

    def _element_constructor_(self, data):
        r"""
        Construct an element out of the given data.

        INPUT:

        - ``data`` -- an object

        OUTPUT: a :class:`Sign`

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import SignGroup
            sage: S = SignGroup()
            sage: S(1)  # indirect doctest
            1
            sage: S(-1)  # indirect doctest
            -1
        """
        if isinstance(data, int) and data == 0:
            raise ValueError('no input specified')

        elif isinstance(data, self.element_class):
            if data.parent() == self:
                return data
            element = data._element_

        else:
            element = data

        return self.element_class(self, element)


class ArgumentGroupFactory(UniqueFactory):
    r"""
    A factory for creating argument groups.

    INPUT:

    - ``data`` -- an object

      The factory will analyze ``data`` and interpret it as
      ``specification`` or ``domain``.

    - ``specification`` -- string

      The following is possible:

      - ``'Signs'`` give the :class:`SignGroup`

      - ``'UU'`` give the :class:`RootsOfUnityGroup`

      - ``'UU_P'``, where ``'P'`` is
        a string representing a SageMath parent which is interpreted as
        ``exponents``

      - ``'Arg_P'``, where ``'P'`` is
        a string representing a SageMath parent which is interpreted as
        ``domain``

    - ``domain`` -- a SageMath parent representing a subset of the complex plane;
      an instance of :class:`ArgumentByElementGroup` will be created with the given
      ``domain``

    - ``exponents`` -- a SageMath parent representing a subset of the reals;
      an instance of :class`UnitCircleGroup` will be created with the given
      ``exponents``

    Exactly one of ``data``, ``specification``, ``exponents`` has to be provided.

    Further keyword parameters will be carried on to the initialization of
    the group.

    EXAMPLES::

        sage: from sage.groups.misc_gps.argument_groups import ArgumentGroup

        sage: ArgumentGroup('UU')                                                       # needs sage.rings.number_field
        Group of Roots of Unity

        sage: # needs sage.rings.number_field
        sage: ArgumentGroup(ZZ)
        Sign Group
        sage: ArgumentGroup(QQ)
        Sign Group
        sage: ArgumentGroup('UU_QQ')
        Group of Roots of Unity
        sage: ArgumentGroup(AA)
        Sign Group

        sage: ArgumentGroup(RR)                                                         # needs sage.rings.number_field
        Sign Group
        sage: ArgumentGroup('Arg_RR')                                                   # needs sage.rings.number_field
        Sign Group
        sage: ArgumentGroup(RIF)
        Sign Group
        sage: ArgumentGroup(RBF)
        Sign Group

        sage: ArgumentGroup(CC)                                                         # needs sage.rings.number_field
        Unit Circle Group with Exponents in
        Real Field with 53 bits of precision modulo ZZ
        sage: ArgumentGroup('Arg_CC')                                                   # needs sage.rings.number_field
        Unit Circle Group with Exponents in
        Real Field with 53 bits of precision modulo ZZ
        sage: ArgumentGroup(CIF)
        Unit Circle Group with Exponents in
        Real Interval Field with 53 bits of precision modulo ZZ
        sage: ArgumentGroup(CBF)
        Unit Circle Group with Exponents in
        Real ball field with 53 bits of precision modulo ZZ

        sage: ArgumentGroup(CyclotomicField(3))                                         # needs sage.rings.number_field
        Unit Circle Group with Argument of Elements in
        Cyclotomic Field of order 3 and degree 2
    """
    def create_key_and_extra_args(self,
                                  data=None,
                                  specification=None,
                                  domain=None,
                                  exponents=None,
                                  **kwds):
        r"""
        Normalize the input.

        See :class:`ArgumentGroupFactory` for a description and examples.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import ArgumentGroup

            sage: # needs sage.rings.number_field
            sage: ArgumentGroup(specification='UU')
            Group of Roots of Unity
            sage: ArgumentGroup('UU') is ArgumentGroup(exponents=QQ)  # indirect doctest
            True
            sage: ArgumentGroup('Arg_CC') is ArgumentGroup(exponents=RR)  # indirect doctest
            True
            sage: ArgumentGroup('Arg_CC') is ArgumentGroup(domain=CC)  # indirect doctest
            True
        """
        from sage.misc.misc import exactly_one_is_true
        from sage.rings.integer_ring import ZZ
        from sage.rings.qqbar import AA
        from sage.rings.rational_field import QQ

        if not exactly_one_is_true(
                (data is not None,
                 specification is not None,
                 domain is not None,
                 exponents is not None)):
            raise ValueError(
                'input ambiguous: ' +
                ', '.join('{}={}'.format(s, v) for s, v in
                          [('data', data), ('specification', specification),
                           ('domain', domain), ('exponents', exponents)]
                          if v is not None))

        if data is not None:
            if isinstance(data, str):
                specification = data
            else:
                domain = data

        if specification is not None:
            if specification == 'UU':
                return (RootsOfUnityGroup, ()), kwds
            if specification == 'Signs':
                return (SignGroup, ()), kwds
            if specification.startswith('UU_'):
                from sage.rings.asymptotic.misc import repr_short_to_parent
                exponents = repr_short_to_parent(specification[3:])
            elif specification.startswith('Arg_') or specification.startswith('arg_'):
                from sage.rings.asymptotic.misc import repr_short_to_parent
                domain = repr_short_to_parent(specification[4:])
            else:
                raise ValueError('unknown specification {}'.format(specification))

        if domain is not None:
            if domain in (ZZ, QQ, AA) \
               or isinstance(domain, (sage.rings.abc.RealField,
                                      sage.rings.abc.RealIntervalField,
                                      sage.rings.abc.RealBallField)):
                return (SignGroup, ()), kwds
            if isinstance(domain, (sage.rings.abc.ComplexField,
                                     sage.rings.abc.ComplexIntervalField,
                                     sage.rings.abc.ComplexBallField)):
                return (UnitCircleGroup, (domain._real_field(),)), kwds
            return (ArgumentByElementGroup, (domain,)), kwds

        if exponents is not None:
            if exponents == ZZ:
                return (SignGroup, ()), kwds
            if exponents == QQ:
                return (RootsOfUnityGroup, ()), kwds
            return (UnitCircleGroup, (exponents,)), kwds

    def create_object(self, version, key, **kwds):
        r"""
        Create an object from the given arguments.

        TESTS::

            sage: from sage.groups.misc_gps.argument_groups import ArgumentGroup
            sage: ArgumentGroup('UU')  # indirect doctest                               # needs sage.rings.number_field
            Group of Roots of Unity
        """
        cls, args = key
        return cls(*args, **kwds)


ArgumentGroup = ArgumentGroupFactory('sage.groups.misc_gps.argument_groups.ArgumentGroup')
r"""
A factory for argument groups.

This is an instance of :class:`ArgumentGroupFactory` whose documentation
provides more details.
"""



r"""
Morphisms of chain complexes

AUTHORS:

- Benjamin Antieau <d.ben.antieau@gmail.com> (2009.06)

- Travis Scrimshaw (2012-08-18): Made all simplicial complexes immutable to
  work with the homset cache.

This module implements morphisms of chain complexes. The input is a dictionary
whose keys are in the grading group of the chain complex and whose values are
matrix morphisms.

EXAMPLES::

    sage: # needs sage.graphs
    sage: S = simplicial_complexes.Sphere(1); S
    Minimal triangulation of the 1-sphere
    sage: C = S.chain_complex()
    sage: C.differential()
    {0: [], 1: [-1 -1  0]
     [ 1  0 -1]
     [ 0  1  1], 2: []}
    sage: f = {0: zero_matrix(ZZ,3,3), 1: zero_matrix(ZZ,3,3)}
    sage: G = Hom(C, C)
    sage: x = G(f); x
    Chain complex endomorphism of
     Chain complex with at most 2 nonzero terms over Integer Ring
    sage: x._matrix_dictionary
    {0: [0 0 0]
        [0 0 0]
        [0 0 0],
     1: [0 0 0]
        [0 0 0]
        [0 0 0]}
"""

# ****************************************************************************
# Copyright (C) 2009 D. Benjamin Antieau <d.ben.antieau@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty
#    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License for more details; the full text
#  is available at:
#
#                  https://www.gnu.org/licenses/
#
# ****************************************************************************

from sage.matrix.constructor import block_diagonal_matrix, zero_matrix
from sage.categories.morphism import Morphism
from sage.categories.homset import Hom
from sage.categories.chain_complexes import ChainComplexes


class ChainComplexMorphism(Morphism):
    """
    An element of this class is a morphism of chain complexes.
    """
    def __init__(self, matrices, C, D, check=True) -> None:
        """
        Create a morphism from a dictionary of matrices.

        EXAMPLES::

            sage: # needs sage.graphs
            sage: S = simplicial_complexes.Sphere(1); S
            Minimal triangulation of the 1-sphere
            sage: C = S.chain_complex()
            sage: C.differential()
            {0: [],
             1: [-1 -1  0]
                [ 1  0 -1]
                [ 0  1  1],
             2: []}
            sage: f = {0:zero_matrix(ZZ,3,3),1:zero_matrix(ZZ,3,3)}
            sage: G = Hom(C,C)
            sage: x = G(f); x
            Chain complex endomorphism of
             Chain complex with at most 2 nonzero terms over Integer Ring
            sage: x._matrix_dictionary
            {0: [0 0 0]
                [0 0 0]
                [0 0 0],
             1: [0 0 0]
                [0 0 0]
                [0 0 0]}

        Check that the bug in :issue:`13220` has been fixed::

            sage: # needs sage.graphs
            sage: X = simplicial_complexes.Simplex(1)
            sage: Y = simplicial_complexes.Simplex(0)
            sage: g = Hom(X,Y)({0: 0, 1: 0})
            sage: g.associated_chain_complex_morphism()
            Chain complex morphism:
              From: Chain complex with at most 2 nonzero terms over Integer Ring
              To: Chain complex with at most 1 nonzero terms over Integer Ring

        Check that an error is raised if the matrices are the wrong size::

            sage: C = ChainComplex({0: zero_matrix(ZZ, 0, 1)})
            sage: D = ChainComplex({0: zero_matrix(ZZ, 0, 2)})
            sage: Hom(C,D)({0: matrix(1, 2, [1, 1])})  # 1x2 is the wrong size.
            Traceback (most recent call last):
            ...
            ValueError: matrix in degree 0 is not the right size
            sage: Hom(C,D)({0: matrix(2, 1, [1, 1])})  # 2x1 is right.
            Chain complex morphism:
              From: Chain complex with at most 1 nonzero terms over Integer Ring
              To: Chain complex with at most 1 nonzero terms over Integer Ring
        """
        if not C.base_ring() == D.base_ring():
            raise NotImplementedError('morphisms between chain complexes of different'
                                      ' base rings are not implemented')
        d = C.degree_of_differential()
        if d != D.degree_of_differential():
            raise ValueError('degree of differential does not match')

        degrees = list(C.differential()) + list(D.differential())
        degrees = sorted(set(degrees))
        initial_matrices = dict(matrices)
        matrices = dict()
        for i in degrees:
            if i - d not in degrees:
                if not (C.free_module_rank(i) == D.free_module_rank(i) == 0):
                    raise ValueError('{} and {} are not rank 0 in degree {}'.format(C, D, i))
                continue
            try:
                matrices[i] = initial_matrices.pop(i)
            except KeyError:
                matrices[i] = zero_matrix(C.base_ring(),
                                          D.differential(i).ncols(),
                                          C.differential(i).ncols(), sparse=True)
        if check:
            # All remaining matrices given must be 0x0.
            if not all(m.ncols() == m.nrows() == 0 for m in initial_matrices.values()):
                raise ValueError('the remaining matrices are not empty')
            # Check sizes of matrices.
            for i in matrices:
                if (matrices[i].nrows() != D.free_module_rank(i) or
                        matrices[i].ncols() != C.free_module_rank(i)):
                    raise ValueError(f'matrix in degree {i} is not the right size')
            # Check commutativity.
            for i in degrees:
                if i - d not in degrees:
                    if not (C.free_module_rank(i) == D.free_module_rank(i) == 0):
                        raise ValueError('{} and {} are not rank 0 in degree {}'.format(C, D, i))
                    continue
                if i + d not in degrees:
                    if not (C.free_module_rank(i+d) == D.free_module_rank(i+d) == 0):
                        raise ValueError('{} and {} are not rank 0 in degree {}'.format(C, D, i+d))
                    continue
                Dm = D.differential(i) * matrices[i]
                mC = matrices[i+d] * C.differential(i)
                if mC != Dm:
                    raise ValueError('matrices must define a chain complex morphism')
        self._matrix_dictionary = {}
        for i in matrices:
            m = matrices[i]
            # Use immutable matrices because they're hashable.
            m.set_immutable()
            self._matrix_dictionary[i] = m
        Morphism.__init__(self, Hom(C, D, ChainComplexes(C.base_ring())))

    def in_degree(self, n):
        """
        The matrix representing this morphism in degree `n`.

        INPUT:

        - ``n`` -- degree

        EXAMPLES::

            sage: C = ChainComplex({0: identity_matrix(ZZ, 1)})
            sage: D = ChainComplex({0: zero_matrix(ZZ, 1), 1: zero_matrix(ZZ, 1)})
            sage: f = Hom(C,D)({0: identity_matrix(ZZ, 1), 1: zero_matrix(ZZ, 1)})
            sage: f.in_degree(0)
            [1]

        Note that if the matrix is not specified in the definition of
        the map, it is assumed to be zero::

            sage: f.in_degree(2)
            []
            sage: f.in_degree(2).nrows(), f.in_degree(2).ncols()
            (1, 0)
            sage: C.free_module(2)
            Ambient free module of rank 0 over the principal ideal domain Integer Ring
            sage: D.free_module(2)
            Ambient free module of rank 1 over the principal ideal domain Integer Ring
        """
        try:
            return self._matrix_dictionary[n]
        except KeyError:
            rows = self.codomain().free_module_rank(n)
            cols = self.domain().free_module_rank(n)
            return zero_matrix(self.domain().base_ring(), rows, cols)

    def to_matrix(self, deg=None):
        """
        The matrix representing this chain map.

        If the degree ``deg`` is specified, return the matrix in that
        degree; otherwise, return the (block) matrix for the whole
        chain map.

        INPUT:

        - ``deg`` -- (default: ``None``) the degree

        EXAMPLES::

            sage: C = ChainComplex({0: identity_matrix(ZZ, 1)})
            sage: D = ChainComplex({0: zero_matrix(ZZ, 1), 1: zero_matrix(ZZ, 1)})
            sage: f = Hom(C,D)({0: identity_matrix(ZZ, 1), 1: zero_matrix(ZZ, 1)})
            sage: f.to_matrix(0)
            [1]
            sage: f.to_matrix()
            [1|0|]
            [-+-+]
            [0|0|]
            [-+-+]
            [0|0|]
        """
        if deg is not None:
            return self.in_degree(deg)
        blocks = [self._matrix_dictionary[n]
                  for n in sorted(self._matrix_dictionary)]
        return block_diagonal_matrix(blocks)

    def dual(self):
        """
        The dual chain map to this one.

        That is, the map from the dual of the codomain of this one to
        the dual of its domain, represented in each degree by the
        transpose of the corresponding matrix.

        EXAMPLES::

            sage: # needs sage.graphs
            sage: X = simplicial_complexes.Simplex(1)
            sage: Y = simplicial_complexes.Simplex(0)
            sage: g = Hom(X,Y)({0:0, 1:0})
            sage: f = g.associated_chain_complex_morphism()
            sage: f.in_degree(0)
            [1 1]
            sage: f.dual()
            Chain complex morphism:
              From: Chain complex with at most 1 nonzero terms over Integer Ring
                To: Chain complex with at most 2 nonzero terms over Integer Ring
            sage: f.dual().in_degree(0)
            [1]
            [1]
            sage: ascii_art(f.domain())
                        [-1]
                        [ 1]
             0 <-- C_0 <----- C_1 <-- 0
            sage: ascii_art(f.dual().codomain())
                        [-1  1]
             0 <-- C_1 <-------- C_0 <-- 0
        """
        matrix_dict = self._matrix_dictionary
        matrices = {i: matrix_dict[i].transpose() for i in matrix_dict}
        return ChainComplexMorphism(matrices, self.codomain().dual(), self.domain().dual())

    def __neg__(self):
        """
        Return ``-x``.

        EXAMPLES::

            sage: # needs sage.graphs
            sage: S = simplicial_complexes.Sphere(2)
            sage: H = Hom(S,S)
            sage: i = H.identity()
            sage: x = i.associated_chain_complex_morphism()
            sage: w = -x
            sage: w._matrix_dictionary
            {0: [-1  0  0  0]
                [ 0 -1  0  0]
                [ 0  0 -1  0]
                [ 0  0  0 -1],
             1: [-1  0  0  0  0  0]
                [ 0 -1  0  0  0  0]
                [ 0  0 -1  0  0  0]
                [ 0  0  0 -1  0  0]
                [ 0  0  0  0 -1  0]
                [ 0  0  0  0  0 -1],
             2: [-1  0  0  0]
                [ 0 -1  0  0]
                [ 0  0 -1  0]
                [ 0  0  0 -1]}
        """
        f = dict()
        for i in self._matrix_dictionary.keys():
            f[i] = -self._matrix_dictionary[i]
        return ChainComplexMorphism(f, self.domain(), self.codomain())

    def __add__(self, x):
        """
        Return ``self + x``.

        EXAMPLES::

            sage: # needs sage.graphs
            sage: S = simplicial_complexes.Sphere(2)
            sage: H = Hom(S,S)
            sage: i = H.identity()
            sage: x = i.associated_chain_complex_morphism()
            sage: z = x+x
            sage: z._matrix_dictionary
            {0: [2 0 0 0]
                [0 2 0 0]
                [0 0 2 0]
                [0 0 0 2],
             1: [2 0 0 0 0 0]
                [0 2 0 0 0 0]
                [0 0 2 0 0 0]
                [0 0 0 2 0 0]
                [0 0 0 0 2 0]
                [0 0 0 0 0 2],
             2: [2 0 0 0]
                [0 2 0 0]
                [0 0 2 0]
                [0 0 0 2]}
        """
        if not isinstance(x, ChainComplexMorphism) or self.codomain() != x.codomain() or self.domain() != x.domain() or self._matrix_dictionary.keys() != x._matrix_dictionary.keys():
            raise TypeError("unsupported operation")
        f = dict()
        for i in self._matrix_dictionary.keys():
            f[i] = self._matrix_dictionary[i] + x._matrix_dictionary[i]
        return ChainComplexMorphism(f, self.domain(), self.codomain())

    def __mul__(self, x):
        """
        Return ``self * x`` if ``self`` and ``x`` are composable morphisms
        or if ``x`` is an element of the base ring.

        EXAMPLES::

            sage: # needs sage.graphs
            sage: S = simplicial_complexes.Sphere(2)
            sage: H = Hom(S,S)
            sage: i = H.identity()
            sage: x = i.associated_chain_complex_morphism()
            sage: y = x*2
            sage: y._matrix_dictionary
            {0: [2 0 0 0]
                [0 2 0 0]
                [0 0 2 0]
                [0 0 0 2],
             1: [2 0 0 0 0 0]
                [0 2 0 0 0 0]
                [0 0 2 0 0 0]
                [0 0 0 2 0 0]
                [0 0 0 0 2 0]
                [0 0 0 0 0 2],
             2: [2 0 0 0]
                [0 2 0 0]
                [0 0 2 0]
                [0 0 0 2]}
            sage: z = y*y
            sage: z._matrix_dictionary
            {0: [4 0 0 0]
                [0 4 0 0]
                [0 0 4 0]
                [0 0 0 4],
             1: [4 0 0 0 0 0]
                [0 4 0 0 0 0]
                [0 0 4 0 0 0]
                [0 0 0 4 0 0]
                [0 0 0 0 4 0]
                [0 0 0 0 0 4],
             2: [4 0 0 0]
                [0 4 0 0]
                [0 0 4 0]
                [0 0 0 4]}

        TESTS:

        Make sure that the product is taken in the correct order
        (``self * x``, not ``x * self`` -- see :issue:`19065`)::

            sage: C = ChainComplex({0: zero_matrix(ZZ, 0, 2)})
            sage: D = ChainComplex({0: zero_matrix(ZZ, 0, 1)})
            sage: f = Hom(C,D)({0: matrix(1, 2, [1, 1])})
            sage: g = Hom(D,C)({0: matrix(2, 1, [1, 1])})
            sage: (f*g).in_degree(0)
            [2]

        Before :issue:`19065`, the following multiplication produced a
        :exc:`KeyError` because `f` was not explicitly defined in degree 2::

            sage: C0 = ChainComplex({0: zero_matrix(ZZ, 0, 1)})
            sage: C1 = ChainComplex({1: zero_matrix(ZZ, 0, 1)})
            sage: C2 = ChainComplex({2: zero_matrix(ZZ, 0, 1)})
            sage: f = ChainComplexMorphism({}, C0, C1)
            sage: g = ChainComplexMorphism({}, C1, C2)
            sage: g * f
            Chain complex morphism:
              From: Chain complex with at most 1 nonzero terms over Integer Ring
              To: Chain complex with at most 1 nonzero terms over Integer Ring
            sage: f._matrix_dictionary
            {0: [], 1: []}
            sage: g._matrix_dictionary
            {1: [], 2: []}
        """
        if not isinstance(x, ChainComplexMorphism) or self.domain() != x.codomain():
            try:
                y = self.domain().base_ring()(x)
            except TypeError:
                raise TypeError("multiplication is not defined")
            f = dict()
            for i in self._matrix_dictionary:
                f[i] = self._matrix_dictionary[i] * y
            return ChainComplexMorphism(f, self.domain(), self.codomain())
        f = {}
        for i in self._matrix_dictionary:
            f[i] = self._matrix_dictionary[i]*x.in_degree(i)
        return ChainComplexMorphism(f, x.domain(), self.codomain())

    def __rmul__(self, x):
        """
        Return ``x * self`` if ``x`` is an element of the base ring.

        EXAMPLES::

            sage: # needs sage.graphs
            sage: S = simplicial_complexes.Sphere(2)
            sage: H = Hom(S,S)
            sage: i = H.identity()
            sage: x = i.associated_chain_complex_morphism()
            sage: 2*x == x*2
            True
            sage: 3*x == x*2
            False
        """
        try:
            y = self.domain().base_ring()(x)
        except TypeError:
            raise TypeError("multiplication is not defined")
        f = dict()
        for i in self._matrix_dictionary:
            f[i] = y * self._matrix_dictionary[i]
        return ChainComplexMorphism(f, self.domain(), self.codomain())

    def __sub__(self, x):
        """
        Return ``self - x``.

        EXAMPLES::

            sage: # needs sage.graphs
            sage: S = simplicial_complexes.Sphere(2)
            sage: H = Hom(S,S)
            sage: i = H.identity()
            sage: x = i.associated_chain_complex_morphism()
            sage: y = x - x
            sage: y._matrix_dictionary
            {0: [0 0 0 0]
            [0 0 0 0]
            [0 0 0 0]
            [0 0 0 0],
             1: [0 0 0 0 0 0]
            [0 0 0 0 0 0]
            [0 0 0 0 0 0]
            [0 0 0 0 0 0]
            [0 0 0 0 0 0]
            [0 0 0 0 0 0],
             2: [0 0 0 0]
            [0 0 0 0]
            [0 0 0 0]
            [0 0 0 0]}
        """
        return self + (-x)

    def __eq__(self, x) -> bool:
        """
        Return ``True`` if and only if ``self == x``.

        EXAMPLES::

            sage: # needs sage.graphs
            sage: S = SimplicialComplex(immutable=True)
            sage: H = Hom(S,S)
            sage: i = H.identity()
            sage: x = i.associated_chain_complex_morphism(); x
            Chain complex morphism:
              From: Trivial chain complex over Integer Ring
              To: Trivial chain complex over Integer Ring
            sage: f = x._matrix_dictionary
            sage: C = S.chain_complex()
            sage: G = Hom(C,C)
            sage: y = G(f)
            sage: x == y
            True
        """
        return isinstance(x, ChainComplexMorphism) \
            and self.codomain() == x.codomain() \
            and self.domain() == x.domain() \
            and self._matrix_dictionary == x._matrix_dictionary

    def is_identity(self) -> bool:
        """
        Return ``True`` if this is the identity map.

        EXAMPLES::

            sage: # needs sage.graphs
            sage: S = SimplicialComplex(immutable=True)
            sage: H = Hom(S,S)
            sage: i = H.identity()
            sage: x = i.associated_chain_complex_morphism()
            sage: x.is_identity()
            True
        """
        return self.to_matrix().is_one()

    def is_surjective(self) -> bool:
        """
        Return ``True`` if this map is surjective.

        EXAMPLES::

            sage: # needs sage.graphs
            sage: S1 = simplicial_complexes.Sphere(1)
            sage: H = Hom(S1, S1)
            sage: flip = H({0:0, 1:2, 2:1})
            sage: flip.associated_chain_complex_morphism().is_surjective()
            True

            sage: # needs sage.graphs
            sage: pt = simplicial_complexes.Simplex(0)
            sage: inclusion = Hom(pt, S1)({0:2})
            sage: inclusion.associated_chain_complex_morphism().is_surjective()
            False
            sage: inclusion.associated_chain_complex_morphism(cochain=True).is_surjective()
            True
        """
        m = self.to_matrix()
        return m.rank() == m.nrows()

    def is_injective(self) -> bool:
        """
        Return ``True`` if this map is injective.

        EXAMPLES::

            sage: # needs sage.graphs
            sage: S1 = simplicial_complexes.Sphere(1)
            sage: H = Hom(S1, S1)
            sage: flip = H({0:0, 1:2, 2:1})
            sage: flip.associated_chain_complex_morphism().is_injective()
            True

            sage: # needs sage.graphs
            sage: pt = simplicial_complexes.Simplex(0)
            sage: inclusion = Hom(pt, S1)({0:2})
            sage: inclusion.associated_chain_complex_morphism().is_injective()
            True
            sage: inclusion.associated_chain_complex_morphism(cochain=True).is_injective()
            False
        """
        return self.to_matrix().right_nullity() == 0

    def __hash__(self) -> int:
        """
        TESTS::

            sage: C = ChainComplex({0: identity_matrix(ZZ, 1)})
            sage: D = ChainComplex({0: zero_matrix(ZZ, 1)})
            sage: f = Hom(C,D)({0: identity_matrix(ZZ, 1), 1: zero_matrix(ZZ, 1)})
            sage: hash(f)  # random
            17
        """
        return hash(self.domain()) ^ hash(self.codomain()) ^ hash(tuple(self._matrix_dictionary.items()))

    def _repr_type(self) -> str:
        """
        EXAMPLES::

            sage: C = ChainComplex({0: identity_matrix(ZZ, 1)})
            sage: D = ChainComplex({0: zero_matrix(ZZ, 1)})
            sage: Hom(C,D)({0: identity_matrix(ZZ, 1), 1: zero_matrix(ZZ, 1)})._repr_type()
            'Chain complex'
        """
        return "Chain complex"



"""
Text in plots
"""
#*****************************************************************************
#       Copyright (C) 2006 Alex Clemesha <clemesha@gmail.com>,
#                          William Stein <wstein@gmail.com>,
#                     2008 Mike Hansen <mhansen@gmail.com>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************
from sage.plot.primitive import GraphicPrimitive
from sage.misc.decorators import options, rename_keyword
from sage.plot.colors import to_mpl_color


class Text(GraphicPrimitive):
    """
    Base class for Text graphics primitive.

    TESTS:

    We test creating some text::

        sage: text("I like Fibonacci",(3,5))
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        sphinx_plot(text("I like Fibonacci",(3,5)))
    """
    def __init__(self, string, point, options):
        """
        Initialize base class Text.

        EXAMPLES::

            sage: T = text("I like Fibonacci", (3,5))
            sage: t = T[0]
            sage: t.string
            'I like Fibonacci'
            sage: t.x
            3.0
            sage: t.options()['fontsize']
            10
        """
        self.string = string
        self.x = float(point[0])
        self.y = float(point[1])
        GraphicPrimitive.__init__(self, options)

    def get_minmax_data(self):
        """
        Return a dictionary with the bounding box data. Notice
        that, for text, the box is just the location itself.

        EXAMPLES::

            sage: T = text("Where am I?",(1,1))
            sage: t=T[0]
            sage: t.get_minmax_data()['ymin']
            1.0
            sage: t.get_minmax_data()['ymax']
            1.0
        """
        from sage.plot.plot import minmax_data
        return minmax_data([self.x], [self.y], dict=True)

    def _repr_(self):
        """
        String representation of Text primitive.

        EXAMPLES::

            sage: T = text("I like cool constants", (pi,e))                             # needs sage.symbolic
            sage: t = T[0];t                                                            # needs sage.symbolic
            Text 'I like cool constants' at the point (3.1415926535...,2.7182818284...)
        """
        return f"Text '{self.string}' at the point ({self.x},{self.y})"

    def _allowed_options(self):
        """
        Return the allowed options for the Text class.

        EXAMPLES::

            sage: T = text("ABC",(1,1),zorder=3)
            sage: T[0]._allowed_options()['fontsize']
            "How big the text is. Either the size in points or a relative size, e.g. 'smaller', 'x-large', etc"
            sage: T[0]._allowed_options()['zorder']
            'The layer level in which to draw'
            sage: T[0]._allowed_options()['rotation']
            'How to rotate the text: angle in degrees, vertical, horizontal'
        """
        return {'fontsize': 'How big the text is. Either the size in points or a relative size, e.g. \'smaller\', \'x-large\', etc',
                'fontstyle': 'A string either \'normal\', \'italic\' or \'oblique\'',
                'fontweight': 'A numeric value in the range 0-1000 or a string'
                              '\'ultralight\', \'light\', \'normal\', \'regular\', \'book\','
                              '\'medium\', \'roman\', \'semibold\', \'demibold\', \'demi\','
                              '\'bold,\', \'heavy\', \'extra bold\', \'black\'',
                'rgbcolor': 'The color as an RGB tuple',
                'background_color': 'The background color',
                'bounding_box': 'A dictionary specifying a bounding box',
                'hue': 'The color given as a hue',
                'alpha': 'A float (0.0 transparent through 1.0 opaque)',
                'axis_coords': 'If True use axis coordinates: (0,0) lower left and (1,1) upper right',
                'rotation': 'How to rotate the text: angle in degrees, vertical, horizontal',
                'vertical_alignment': 'How to align vertically: top, center, bottom',
                'horizontal_alignment': 'How to align horizontally: left, center, right',
                'zorder': 'The layer level in which to draw',
                'clip': 'Whether to clip or not'}

    def _plot3d_options(self, options=None):
        """
        Translate 2D plot options into 3D plot options.

        EXAMPLES::

            sage: T = text("ABC",(1,1))
            sage: t = T[0]
            sage: t.options()['rgbcolor']
            (0.0, 0.0, 1.0)
            sage: s=t.plot3d()
            sage: s.jmol_repr(s.testing_render_params())[0][1]
            'color atom  [0,0,255]'
        """
        if options is None:
            options = dict(self.options())
        options_3d = {}
        for s in ['fontfamily', 'fontsize', 'fontstyle', 'fontweight']:
            if s in options:
                options_3d[s] = options.pop(s)
        # TODO: figure out how to implement rather than ignore
        for s in ['axis_coords', 'clip', 'horizontal_alignment',
                  'rotation', 'vertical_alignment']:
            if s in options:
                del options[s]
        options_3d.update(GraphicPrimitive._plot3d_options(self, options))
        return options_3d

    def plot3d(self, **kwds):
        """
        Plot 2D text in 3D.

        EXAMPLES::

            sage: T = text("ABC", (1, 1))
            sage: t = T[0]
            sage: s = t.plot3d()
            sage: s.jmol_repr(s.testing_render_params())[0][2]
            'label "ABC"'
            sage: s._trans
            (1.0, 1.0, 0)
        """
        from sage.plot.plot3d.shapes2 import text3d
        options = self._plot3d_options()
        options.update(kwds)
        return text3d(self.string, (self.x, self.y, 0), **options)

    def _render_on_subplot(self, subplot):
        """
        TESTS::

            sage: t1 = text("Hello", (1,1), vertical_alignment='top', fontsize=30, rgbcolor='black')
            sage: t2 = text("World", (1,1), horizontal_alignment='left', fontsize=20, zorder=-1)
            sage: t1 + t2   # render the sum
            Graphics object consisting of 2 graphics primitives
        """
        options = self.options()
        opts = {}
        opts['color'] = options['rgbcolor']
        opts['verticalalignment'] = options['vertical_alignment']
        opts['horizontalalignment'] = options['horizontal_alignment']
        if 'background_color' in options:
            opts['backgroundcolor'] = options['background_color']
        if 'fontweight' in options:
            opts['fontweight'] = options['fontweight']
        if 'alpha' in options:
            opts['alpha'] = options['alpha']
        if 'fontstyle' in options:
            opts['fontstyle'] = options['fontstyle']
        if 'bounding_box' in options:
            opts['bbox'] = options['bounding_box']
        if 'zorder' in options:
            opts['zorder'] = options['zorder']
        if options['axis_coords']:
            opts['transform'] = subplot.transAxes
        if 'fontsize' in options:
            val = options['fontsize']
            if isinstance(val, str):
                opts['fontsize'] = val
            else:
                opts['fontsize'] = int(val)
        if 'rotation' in options:
            val = options['rotation']
            if isinstance(val, str):
                opts['rotation'] = options['rotation']
            else:
                opts['rotation'] = float(options['rotation'])

        p = subplot.text(self.x, self.y, self.string, clip_on=options['clip'], **opts)
        if not options['clip']:
            self._bbox_extra_artists = [p]


@rename_keyword(color='rgbcolor')
@options(fontsize=10, rgbcolor=(0,0,1), horizontal_alignment='center',
         vertical_alignment='center', axis_coords=False, clip=False)
def text(string, xy, **options):
    r"""
    Return a 2D text graphics object at the point `(x, y)`.

    Type ``text.options`` for a dictionary of options for 2D text.

    2D OPTIONS:

    - ``fontsize`` -- how big the text is. Either an integer that
      specifies the size in points or a string which specifies a size (one of
      ``'xx-small'``, ``'x-small'``, ``'small'``, ``'medium'``, ``'large'``,
      ``'x-large'``, ``'xx-large'``).

    - ``fontstyle`` -- string either ``'normal'``, ``'italic'`` or ``'oblique'``

    - ``fontweight`` -- a numeric value in the range 0-1000 or a string (one of
      ``'ultralight'``, ``'light'``, ``'normal'``, ``'regular'``, ``'book'``,
      ``'medium'``, ``'roman'``, ``'semibold'``, ``'demibold'``, ``'demi'``,
      ``'bold'``, ``'heavy'``, ``'extra bold'``, ``'black'``)

    - ``rgbcolor`` -- the color as an RGB tuple

    - ``hue`` -- the color given as a hue

    - ``alpha`` -- a float (0.0 transparent through 1.0 opaque)

    - ``background_color`` -- the background color

    - ``rotation`` -- how to rotate the text: angle in degrees, vertical, horizontal

    - ``vertical_alignment`` -- how to align vertically: top, center, bottom

    - ``horizontal_alignment`` -- how to align horizontally: left, center, right

    - ``zorder`` -- the layer level in which to draw

    - ``clip`` -- boolean (default: ``False``); whether to clip or not

    - ``axis_coords`` -- boolean (default: ``False``); if ``True``, use axis
      coordinates, so that (0,0) is the lower left and (1,1) upper right,
      regardless of the x and y range of plotted values

    - ``bounding_box`` -- dictionary specifying a bounding box; currently the
      text location

    EXAMPLES::

        sage: text("Sage graphics are really neat because they use matplotlib!", (2,12))
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        t = "Sage graphics are really neat because they use matplotlib!"
        sphinx_plot(text(t,(2,12)))

    Larger font, bold, colored red and transparent text::

        sage: text("I had a dream!", (2,12), alpha=0.3,
        ....:      fontsize='large', fontweight='bold', color='red')
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        sphinx_plot(text("I had a dream!", (2,12), alpha=0.3, fontsize='large', fontweight='bold', color='red'))

    By setting ``horizontal_alignment`` to 'left' the text is guaranteed to be
    in the lower left no matter what::

        sage: text("I got a horse and he lives in a tree", (0,0),
        ....:      axis_coords=True, horizontal_alignment='left')
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        t = "I got a horse and he lives in a tree"
        sphinx_plot(text(t, (0,0), axis_coords=True, horizontal_alignment='left'))

    Various rotations::

        sage: text("noitator", (0,0), rotation=45.0,
        ....:      horizontal_alignment='left', vertical_alignment='bottom')
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        sphinx_plot(text("noitator", (0,0), rotation=45.0, horizontal_alignment='left', vertical_alignment='bottom'))

    ::

        sage: text("Sage is really neat!!", (0,0), rotation='vertical')
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        sphinx_plot(text("Sage is really neat!!", (0,0), rotation='vertical'))

    You can also align text differently::

        sage: t1 = text("Hello", (1,1), vertical_alignment='top')
        sage: t2 = text("World", (1,0.5), horizontal_alignment='left')
        sage: t1 + t2   # render the sum
        Graphics object consisting of 2 graphics primitives

    .. PLOT::

        t1 = text("Hello", (1,1), vertical_alignment='top')
        t2 = text("World", (1,0.5), horizontal_alignment='left')
        sphinx_plot(t1 + t2)

    You can save text as part of PDF output::

        sage: import tempfile
        sage: with tempfile.NamedTemporaryFile(suffix='.pdf') as f:
        ....:     text("sage", (0,0), rgbcolor=(0,0,0)).save(f.name)

    Some examples of bounding box::

        sage: bbox = {'boxstyle': "rarrow,pad=0.3", 'fc': "cyan", 'ec': "b", 'lw': 2}
        sage: text("I feel good", (1,2), bounding_box=bbox)
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

         bbox = {'boxstyle':"rarrow,pad=0.3", 'fc':"cyan", 'ec':"b", 'lw':2}
         sphinx_plot(text("I feel good", (1,2), bounding_box=bbox))

    ::

        sage: text("So good", (0,0), bounding_box={'boxstyle': 'round', 'fc': 'w'})
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        bbox = {'boxstyle':'round', 'fc':'w'}
        sphinx_plot(text("So good", (0,0), bounding_box=bbox))

    The possible options of the bounding box are 'boxstyle' (one of 'larrow',
    'rarrow', 'round', 'round4', 'roundtooth', 'sawtooth', 'square'), 'fc' or
    'facecolor', 'ec' or 'edgecolor', 'ha' or 'horizontalalignment', 'va' or
    'verticalalignment', 'lw' or 'linewidth'.

    A text with a background color::

        sage: text("So good", (-2,2), background_color='red')
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        sphinx_plot(text("So good", (-2,2), background_color='red'))

    Use dollar signs for LaTeX and raw strings to avoid having to
    escape backslash characters::

        sage: A = arc((0, 0), 1, sector=(0.0, RDF.pi()))
        sage: a = sqrt(1./2.)
        sage: PQ = point2d([(-a, a), (a, a)])
        sage: botleft = dict(horizontal_alignment='left', vertical_alignment='bottom')
        sage: botright = dict(horizontal_alignment='right', vertical_alignment='bottom')
        sage: tp = text(r'$z_P = e^{3i\pi/4}$',
        ....:           (-a, a), **botright)
        sage: tq = text(r'$Q = (\frac{\sqrt{2}}{2}, \frac{\sqrt{2}}{2})$',
        ....:           (a, a), **botleft)
        sage: A + PQ + tp + tq
        Graphics object consisting of 4 graphics primitives

    .. PLOT::

        A = arc((0, 0), 1, sector=(0.0, RDF.pi()))
        a = sqrt(1./2.)
        PQ = point2d([(-a, a), (a, a)])
        botleft = dict(horizontal_alignment='left', vertical_alignment='bottom')
        botright = dict(horizontal_alignment='right', vertical_alignment='bottom')
        tp = text(r'$z_P = e^{3i\pi/4}$', (-a, a), **botright)
        tq = text(r'$Q = (\frac{\sqrt{2}}{2}, \frac{\sqrt{2}}{2})$', (a, a), **botleft)
        sphinx_plot(A + PQ + tp + tq)

    Text coordinates must be 2D, an error is raised if 3D coordinates are passed::

        sage: t = text("hi", (1, 2, 3))
        Traceback (most recent call last):
        ...
        ValueError: use text3d instead for text in 3d

    Use the :func:`text3d <sage.plot.plot3d.shapes2.text3d>` function for 3D text::

        sage: t = text3d("hi", (1, 2, 3))

    Or produce 2D text with coordinates `(x, y)` and plot it in 3D (at `z = 0`)::

        sage: t = text("hi", (1, 2))
        sage: t.plot3d()  # text at position (1, 2, 0)
        Graphics3d Object

    Extra options will get passed on to ``show()``, as long as they are valid. Hence this ::

        sage: text("MATH IS AWESOME", (0, 0), fontsize=40, axes=False)
        Graphics object consisting of 1 graphics primitive

    is equivalent to ::

        sage: text("MATH IS AWESOME", (0, 0), fontsize=40).show(axes=False)
    """
    try:
        x, y = xy
    except ValueError:
        if isinstance(xy, (list, tuple)) and len(xy) == 3:
            raise ValueError("use text3d instead for text in 3d")
        raise
    from sage.plot.graphics import Graphics
    options['rgbcolor'] = to_mpl_color(options['rgbcolor'])
    point = (float(x), float(y))
    g = Graphics()
    g._set_extra_kwds(Graphics._extract_kwds_for_show(options, ignore='fontsize'))
    g.add_primitive(Text(string, point, options))
    return g



# sage.doctest: needs sage.rings.finite_rings
r"""

Cyclic covers over a finite field

The most interesting feature is computation of Frobenius matrix on
Monsky-Washnitzer cohomology and the Frobenius polynomial.

REFERENCES:

- [ABCMT2019]_

EXAMPLES::

    sage: p = 13
    sage: x = PolynomialRing(GF(p),"x").gen()
    sage: C = CyclicCover(4, x^4 + 1)
    sage: C.frobenius_polynomial()
    x^6 - 6*x^5 + 3*x^4 + 60*x^3 + 39*x^2 - 1014*x + 2197
    sage: R.<t> = PowerSeriesRing(Integers())
    sage: C.projective_closure().zeta_series(2,t)
    1 + 8*t + 102*t^2 + O(t^3)
    sage: C.frobenius_polynomial().reverse()(t)/((1-t)*(1-p*t)) + O(t^5)
    1 + 8*t + 102*t^2 + 1384*t^3 + 18089*t^4 + O(t^5)

    sage: p = 49999
    sage: x = PolynomialRing(GF(p),"x").gen()
    sage: CyclicCover(5, x^5 + x).frobenius_polynomial()  # long time
    x^12 + 299994*x^10 + 37498500015*x^8 + 2499850002999980*x^6 + 93742500224997000015*x^4
     + 1874812507499850001499994*x^2 + 15623125093747500037499700001
    sage: CyclicCover(5, 2*x^5 + x).frobenius_polynomial()  # long time
    x^12 + 299994*x^10 + 37498500015*x^8 + 2499850002999980*x^6 + 93742500224997000015*x^4
     + 1874812507499850001499994*x^2 + 15623125093747500037499700001

    sage: p = 107
    sage: x = PolynomialRing(GF(p),"x").gen()
    sage: CyclicCover(2, x^5 + x).frobenius_matrix()
    [              O(107^2)      89*107 + O(107^2)               O(107^2)               O(107^2)]
    [     89*107 + O(107^2)               O(107^2)               O(107^2)               O(107^2)]
    [              O(107^2)               O(107^2)               O(107^2) 105 + 5*107 + O(107^2)]
    [              O(107^2)               O(107^2) 89 + 53*107 + O(107^2)               O(107^2)]
    sage: CyclicCover(2, 3*x^5 + x).frobenius_matrix()
    [              O(107^2)      14*107 + O(107^2)               O(107^2)               O(107^2)]
    [     69*107 + O(107^2)               O(107^2)               O(107^2)               O(107^2)]
    [              O(107^2)               O(107^2)               O(107^2) 61 + 58*107 + O(107^2)]
    [              O(107^2)               O(107^2) 69 + 53*107 + O(107^2)               O(107^2)]
    sage: CyclicCover(3, x^3 + x).frobenius_matrix()
    [          0           0      O(107)      O(107)]
    [          0           0 52 + O(107)      O(107)]
    [     O(107) 35 + O(107)           0           0]
    [44 + O(107)      O(107)           0           0]
    sage: CyclicCover(3, 3*x^3 + x).frobenius_matrix()
    [          0           0      O(107)      O(107)]
    [          0           0 79 + O(107)      O(107)]
    [     O(107) 42 + O(107)           0           0]
    [30 + O(107)      O(107)           0           0]
"""

# *****************************************************************************
#  Copyright (C) 2018   Vishal Arul <varul@mit.edu>,
#                       Alex Best <alex.j.best@gmail.com>,
#                       Edgar Costa <edgarc@mit.edu>,
#                       Richard Magner <rmagner@bu.edu>,
#                       Nicholas Triantafillou <ngtriant@mit.edu>
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# *****************************************************************************

from sage.arith.misc import euler_phi
from sage.matrix.constructor import matrix, zero_matrix
from sage.misc.cachefunc import cached_method
from sage.misc.lazy_import import lazy_import
from sage.modules.free_module_element import vector
from sage.rings.finite_rings.integer_mod_ring import IntegerModRing
from sage.rings.integer_ring import ZZ
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.rings.power_series_ring import PowerSeriesRing

lazy_import("sage.functions.log", "log")
lazy_import("sage.functions.other", ["ceil", "binomial", "floor"])
lazy_import('sage.rings.padics.factory', ['Zp', 'Zq', 'Qq'])
lazy_import('sage.schemes.hyperelliptic_curves.hypellfrob', 'interval_products')

from .charpoly_frobenius import charpoly_frobenius
from . import cycliccover_generic


def _N0_nodenominators(p, g, n):
    """
    Return the necessary `p`-adic precision for the Frobenius matrix to deduce
    the characteristic polynomial of Frobenius using the Newton identities,
    using :meth:`charpoly_frobenius`, which assumes that the Frobenius matrix
    is integral, i.e., has no denominators.

    INPUT:

    - ``p`` -- prime
    - ``g`` -- genus
    - ``n`` -- degree of residue field

    TESTS::

        sage: sage.schemes.cyclic_covers.cycliccover_finite_field._N0_nodenominators(4999, 4, 5)
        11
    """
    return max(
        ceil(log(2 * (2 * g) / ZZ(i), p) + (n * i) / ZZ(2)) for i in range(1, g + 1)
    )


class CyclicCover_finite_field(cycliccover_generic.CyclicCover_generic):
    def __init__(self, AA, r, f, names=None, verbose=0):
        """
        EXAMPLES::

            sage: p = 13
            sage: x = PolynomialRing(GF(p), "x").gen()
            sage: C = CyclicCover(4, x^4 + 1)
            sage: C.frobenius_polynomial()
            x^6 - 6*x^5 + 3*x^4 + 60*x^3 + 39*x^2 - 1014*x + 2197
            sage: R.<t> = PowerSeriesRing(Integers())
            sage: C.projective_closure().zeta_series(2,t)
            1 + 8*t + 102*t^2 + O(t^3)
            sage: C.frobenius_polynomial().reverse()(t)/((1-t)*(1-p*t)) + O(t^5)
            1 + 8*t + 102*t^2 + 1384*t^3 + 18089*t^4 + O(t^5)
        """
        cycliccover_generic.CyclicCover_generic.__init__(self, AA, r, f, names=names)
        self._verbose = verbose
        self._init_frobQ = False
        self._N0 = None

    def _init_frob(self, desired_prec=None):
        """
        Initialise everything for Frobenius polynomial computation.

        TESTS::

            sage: p = 4999
            sage: x = PolynomialRing(GF(p),"x").gen()
            sage: C = CyclicCover(3, x^4 + 4*x^3 + 9*x^2 + 3*x + 1)
            sage: C._init_frob()
            sage: C._init_frobQ
            True
            sage: C._plarge
            True
            sage: C._sqrtp
            True
        """

        def _N0_RH():
            return ceil(
                log(2 * binomial(2 * self._genus, self._genus), self._p)
                + self._genus * self._n / ZZ(2)
            )

        def _find_N0():
            if self._nodenominators:
                return _N0_nodenominators(self._p, self._genus, self._n)
            return _N0_RH() + self._extraprec

        def _find_N_43():
            """
            Find the precision used for thm 4.3 in Goncalves
            for p >> 0, N = N0 + 2
            """
            p = self._p
            r = self._r
            d = self._d
            delta = self._delta
            N0 = self._N0
            left_side = N0 + floor(log((d * p * (r - 1) + r) / delta) / log(p))

            def right_side_log(n):
                return floor(log(p * (r * n - 1) - r) / log(p))

            n = left_side
            while n <= left_side + right_side_log(n):
                n += 1
            return n

        if not self._init_frobQ or self._N0 != desired_prec:
            if self._r < 2 or self._d < 2:
                raise NotImplementedError("Only implemented for r, f.degree() >= 2")

            self._init_frobQ = True

            self._Fq = self._f.base_ring()
            self._p = self._Fq.characteristic()
            self._q = self._Fq.cardinality()
            self._n = self._Fq.degree()
            self._epsilon = 0 if self._delta == 1 else 1

            # our basis choice doesn't always give an integral matrix
            if self._epsilon == 0:
                self._extraprec = floor(
                    log(self._r, self._p)
                    + log((2 * self._genus + (self._delta - 2)) / self._delta, self._p)
                )
            else:
                self._extraprec = floor(log(self._r * 2 - 1, self._p))

            self._nodenominators = self._extraprec == 0

            if desired_prec is None:
                self._N0 = _find_N0()
            else:
                self._N0 = desired_prec

            self._plarge = self._p > self._d * self._r * (self._N0 + self._epsilon)

            # working prec
            if self._plarge:
                self._N = self._N0 + 1
            else:
                self._N = _find_N_43()

            # we will use the sqrt(p) version?
            self._sqrtp = self._plarge and self._p == self._q
            self._extraworkingprec = self._extraprec
            if not self._plarge:
                # we might have some denominators showing up during horizontal
                # and vertical reductions
                self._extraworkingprec += 2 * ceil(
                    log(self._d * self._r * (self._N0 + self._epsilon), self._p)
                )

            # Rings
            if self._plarge and self._nodenominators:
                if self._n == 1:
                    # IntegerModRing is significantly faster than Zq
                    self._Zq = IntegerModRing(self._p**self._N)
                    if self._sqrtp:
                        self._Zq0 = IntegerModRing(self._p**(self._N - 1))
                    self._Qq = Qq(self._p, prec=self._N, type='capped-rel')
                    self._w = 1
                else:
                    self._Zq = Zq(
                        self._q,
                        names='w',
                        modulus=self._Fq.polynomial(),
                        prec=self._N,
                        type='capped-abs',
                    )
                    self._w = self._Zq.gen()
                    self._Qq = self._Zq.fraction_field()
            else:
                self._Zq = Qq(
                    self._q,
                    names='w',
                    modulus=self._Fq.polynomial(),
                    prec=self._N + self._extraworkingprec,
                )
                self._w = self._Zq.gen()
                self._Qq = self._Zq
            self._Zp = Zp(self._p, prec=self._N + self._extraworkingprec)

            self._Zqx = PolynomialRing(self._Zq, "x")

            # Want to take a lift of f from Fq to Zq
            if self._n == 1:
                # When n = 1, can lift from Fp[x] to Z[x] and then to Zp[x]
                self._flift = self._Zqx([elt.lift() for elt in self._f.list()])
                self._frobf = self._Zqx(self._flift.list())
            else:  # When n > 1, need to be more careful with the lift
                self._flift = self._Zqx(
                    [
                        elt.polynomial().change_ring(ZZ)(self._Zq.gen())
                        for elt in self._f.list()
                    ]
                )

                self._frobf = self._Zqx([elt.frobenius() for elt in self._flift.list()])

            self._dflift = self._flift.derivative()

            # Set up local cache for Frob(f)^s

            # This variable will store the powers of frob(f)
            frobpow = [None] * (self._N0 + 2)
            frobpow[0] = self._Zqx(1)
            for k in range(self._N0 + 1):
                frobpow[k + 1] = self._frobf * frobpow[k]
            # We don't make it a polynomials as we need to keep track that the
            # ith coefficient represents  (i*p)-th
            self._frobpow_list = [elt.list() for elt in frobpow]

            if self._sqrtp:
                # precision of self._Zq0
                N = self._N - 1
                vandermonde = matrix(self._Zq0, N, N)
                for i in range(N):
                    vandermonde[i, 0] = 1
                    for j in range(1, N):
                        vandermonde[i, j] = vandermonde[i, j - 1] * (i + 1)
                self._vandermonde = vandermonde.inverse()

                self._horizontal_fat_s = {}
                self._vertical_fat_s = {}

    def _divide_vector(self, D, vect, R):
        """
        Divide the vector `vect` by `D` as a vector over `R`.

        TESTS::

            sage: p = 4999
            sage: x = PolynomialRing(GF(p), "x").gen()
            sage: C = CyclicCover(3, x^4 + 4*x^3 + 9*x^2 + 3*x + 1)
            sage: C._init_frob()
            sage: C._divide_vector(p, vector(C._Qq, [p, p^2, p^3]), C._Qq)
            (1 + O(4999^3), 4999 + O(4999^4), 4999^2 + O(4999^5))
        """
        DQq = self._Qq(D).lift_to_precision(self._Qq.precision_cap())
        m = 1 / DQq
        if not R.is_field():
            vectQq = vector(
                self._Qq,
                [
                    m * self._Qq(elt).lift_to_precision(self._Qq.precision_cap())
                    for elt in vect
                ],
            )
            return vector(R, [R(elt) for elt in vectQq])
        return vector(R, [(m * elt).lift_to_precision() for elt in vect])

    def _frob_sparse(self, i, j, N0):
        r"""
        Compute `Frob(x^i y^(-j) dx ) / dx` for y^r = f(x) with N0 terms.

        INPUT:

        - ``i`` -- the power of x in the expression `Frob(x^i dx/y^j) / dx`

        - ``j`` -- the (negative) power of y in the expression
          `Frob(x^i dx/y^j) / dx`

        OUTPUT:

        ``frobij`` -- a Matrix of size  (d * (N0 - 1) + ) x (N0)
                      that represents the Frobenius expansion of
                      x^i dx/y^j modulo p^(N0 + 1)

                      the entry (l, s) corresponds to the coefficient associated
                      to the monomial x**(p * (i + 1 + l) -1) * y**(p * -(j + r*s))
                      (l, s) --> (p * (i + 1 + l) -1, p * -(j + r*s))

        ALGORITHM:

        Compute:

        Frob(x^i dx/y^j) / dx = p * x ** (p * (i+1) - 1) * y ** (-j*p) * Sigma

        where:

        .. MATH::

            Sigma = \sum_{k = 0} ^{N0-1}
                            \sum_{s = 0} ^k
                                (-1) ** (k-s) * binomial(k, s)
                                * binomial(-j/r, k)
                                * self._frobpow[s]
                                * self._y ** (-self._r * self._p * s)
                    = \sum_{s = 0} ^{N0 - 1}
                        \sum_{k = s} ^N0
                            (-1) ** (k-s) * binomial(k, s)
                            * binomial(-j/self._r, k)
                            * self._frobpow[s]
                            * self._y ** (-self._r*self._p*s)
                    = \sum_{s = 0} ^{N0-1}
                            D_{j, s}
                            * self._frobpow[s]
                            * self._y ** (-self._r * self._p * s)
                    = \sum_{s = 0} ^N0
                        \sum_{l = 0} ^(d*s)
                            D_{j, s} * self._frobpow[s][l]
                            * x ** (self._p ** l)
                            * y ** (-self._r * self._p ** s)

        and:

        .. MATH::

            D_{j, s} = \sum_{k = s} ^N0 (-1) ** (k-s) * binomial(k, s) * binomial(-j/self._r, k) )

        TESTS::

            sage: p = 499
            sage: x = PolynomialRing(GF(p), "x").gen()
            sage: C = CyclicCover(3, x^4 + 4*x^3 + 9*x^2 + 3*x + 1)
            sage: C._init_frob()
            sage: C._frob_sparse(2, 0, 1)
            [499]
            sage: C._frob_sparse(2, 0, 2)
            [499   0]
            [  0   0]
            [  0   0]
            [  0   0]
            [  0   0]
            sage: C._frob_sparse(2, 1, 1)
            [499]
            sage: C._frob_sparse(2, 1, 2)
            [ 82834998  41417000]
            [        0 124251000]
            [        0 124250002]
            [        0  41416501]
            [        0  41417000]
            sage: C._frob_sparse(2, 2, 1)
            [499]
        """

        def _extend_frobpow(power):
            if power < len(self._frobpow_list):
                pass
            else:
                frobpow = self._Zqx(self._frobpow_list[-1])
                for k in range(len(self._frobpow_list), power + 1):
                    frobpow *= self._frobf
                self._frobpow_list.extend([frobpow.list()])
            assert power < len(self._frobpow_list)

        _extend_frobpow(N0)
        r = self._r
        Dj = [
            self._Zq(
                sum(
                    [
                        (-1) ** (k - l) * binomial(k, l) * binomial(-ZZ(j) / r, k)
                        for k in range(l, N0)
                    ]
                )
            )
            for l in range(N0)
        ]
        frobij = matrix(self._Zq, self._d * (N0 - 1) + 1, N0)
        for s in range(N0):
            for l in range(self._d * s + 1):
                frobij[l, s] = self._p * Dj[s] * self._frobpow_list[s][l]
        return frobij

    def _horizontal_matrix_reduction(self, s):
        r"""
        Return the tuple of tuples that represents the horizontal matrix
        reduction at pole order ``s``.

        INPUT:

        - ``s`` -- integer

        OUTPUT:

        A tuple of tuples ``( (D0, D1), (M0, M1) )``
        where `MH_{e, s}  = M0 + e * M1` and `DH_{e,s} = D0 + e * D1`

        ALGORITHM:

        Let `W_{e, s}` to be the Qq-vector space of differential forms
        of the form:

            .. MATH:: G x^e y^{-s} dx

        where `\deg G \leq d - 1`.

        Let `v = [G_0, ..., G_{d-1}]` represent G

        There is a map:

            `MH_{e, s} : W_{e, s} \to W_{e-1, s}`

        and a function to:

            `DH: \NN \times \NN \to  Qq`

        such that:

            `G x^e y^{-s} dx \cong H x^{e - 1} y^{-s} dx`

        where `H = DH(e, s)^{-1} * MH_{e,s} ( G )`

        The matrix `MH_{e, s}` can be written as:

             `MH_{e, s}  = M0_{s} + e * M1_{s}`

        similarly:

            `DH_{e,s} = D0_{s} + e * D1_{s}`

        TESTS::

            sage: p = 4999
            sage: x = PolynomialRing(GF(p), "x").gen()
            sage: C = CyclicCover(3, x^4 + 4*x^3 + 9*x^2 + 3*x + 1)
            sage: C._init_frob()
            sage: C._horizontal_matrix_reduction(24995)
            ((99968, 124925014996),
            (
            [           0            0            0            0]
            [       99968            0            0 124924940023]
            [           0        99968            0 124924565143]
            [           0            0        99968 124924715095],
            <BLANKLINE>
            [           0            0            0            3]
            [124925014996            0            0            9]
            [           0 124925014996            0           27]
            [           0            0 124925014996           12]
            ))
            sage: C._horizontal_matrix_reduction(4999)
            ((19984, 124925014996),
            (
            [           0            0            0            0]
            [       19984            0            0 124925000011]
            [           0        19984            0 124924925071]
            [           0            0        19984 124924955047],
            <BLANKLINE>
            [           0            0            0            3]
            [124925014996            0            0            9]
            [           0 124925014996            0           27]
            [           0            0 124925014996           12]
            ))
        """

        f_co = self._flift.list()
        # DH_{e,s} = ((s -r)*d - e * r) * f_d
        m1 = -1 * self._r * f_co[-1]  # r is a, g_co[-1] is lambda
        m0 = (s - self._r) * self._d * f_co[-1]  # j is a*t + beta

        M1 = matrix(
            self._Zq,
            self._d,
            lambda m, n: m1
            if m == n + 1
            else self._r * f_co[m]
            if n == self._d - 1
            else 0,
        )
        M0 = matrix(
            self._Zq,
            self._d,
            lambda m, n: m0
            if m == n + 1
            else (self._r - s) * m * f_co[m]
            if n == self._d - 1
            else 0,
        )

        return ((m0, m1), (M0, M1))

    def _vertical_matrix_reduction(self, s0):
        r"""
        Return the tuple of tuples that represents the vertical matrix reduction.

        OUTPUT:

        A tuple of tuples ``( (D0, D1), (M0, M1) )``
        where MV_t  = M0 + t * M1 and DV_t = D0 + t * D1

        TESTS::

            sage: p = 4999
            sage: x = PolynomialRing(GF(p), "x").gen()
            sage: C = CyclicCover(3, x^4 + 4*x^3 + 9*x^2 + 3*x + 1)
            sage: C._init_frob()
            sage: C._vertical_matrix_reduction(1)
            ((-2, 3),
            (
            [117410728377  65750007895  58235721278]
            [ 67628579544  59175007105  15028573234]
            [ 86414296088            0  27239288985],
            <BLANKLINE>
            [ 51660720493  75142866164            0]
            [ 74203580341   2817857481  75142866164]
            [108017870113            0   2817857481]
            ))
            sage: C._vertical_matrix_reduction(2)
            ((-1, 3),
            (
            [92989296875  7514286617 58235721278]
            [50721434658 60114292932 81717866955]
            [80778581126           0 28178574812],
            <BLANKLINE>
            [ 51660720493  75142866164            0]
            [ 74203580341   2817857481  75142866164]
            [108017870113            0   2817857481]
            ))
        """

        d = self._d
        f_co = (
            [0 for i in range(d - 2)] + self._flift.list() + [0 for i in range(d - 1)]
        )
        fd_co = (
            [0 for i in range(d - 1)] + self._dflift.list() + [0 for i in range(d)]
        )

        rows = [f_co[d - 2 - i:-i - 1] for i in range(d - 1)]
        rows += [fd_co[d - 1 - i:-i - 1] for i in range(d)]

        m = matrix(rows).transpose().inverse()

        a_foo = m[0:d, 0:d]
        b_foo = m[d - 1:2 * d - 1, 0:d]
        a_foo = matrix(d, d, lambda i, j: 1 if i == j and i != d - 1 else 0) * a_foo
        foo = matrix(d, d, lambda i, j: j if i == j - 1 else 0)
        bp_foo = foo * b_foo
        A_vert = a_foo.submatrix(0, 0, d - 1, d - 1)
        Bd_vert = bp_foo.submatrix(0, 0, d - 1, d - 1)
        M1 = (s0 - self._r) * A_vert + self._r * Bd_vert
        M2 = self._r * A_vert
        m1 = s0 - self._r
        m2 = self._r
        return ((m1, m2), (M1, M2))

    def _reduce_vector_horizontal(self, G, e, s, k=1):
        r"""
        INPUT:

        - a vector -- `G \in W_{e, s}`

        OUTPUT:

        - a vector -- `H \in W_{e - k, s}` such that
            `G x^e y^{-s} dx \cong H x^{e - k} y^{-s} dx`

        TESTS::

            sage: p = 4999
            sage: x = PolynomialRing(GF(p), "x").gen()
            sage: C = CyclicCover(3, x^4 + 4*x^3 + 9*x^2 + 3*x + 1)
            sage: C._init_frob()
            sage: C._initialize_fat_horizontal(p, 3)
            sage: C._reduce_vector_horizontal((83283349998, 0, 0, 0), 2*p  - 1, p, p)
            (23734897071, 84632332850, 44254975407, 23684517017)
            sage: C._reduce_vector_horizontal((98582524551, 3200841460, 6361495378, 98571346457), 2*p - 1, p, p)
            (96813533420, 61680190736, 123292559950, 96786566978)
        """
        if self._sqrtp and k == self._p:
            vect = self._reduce_vector_horizontal_BSGS(G, e, s)
        else:
            vect = self._reduce_vector_horizontal_plain(G, e, s, k)
        return vect

    def _reduce_vector_horizontal_BSGS(self, G, e, s):
        r"""
        INPUT:

        - a vector -- `G \in W_{e, s}`

        OUTPUT:

        - a vector -- `H \in W_{e - p, s}` such that
            `G x^e y^{-s} dx \cong H x^{e - p} y^{-s} dx`

        TESTS::

            sage: p = 4999
            sage: x = PolynomialRing(GF(p), "x").gen()
            sage: C = CyclicCover(3, x^4 + 4*x^3 + 9*x^2 + 3*x + 1)
            sage: C._init_frob()
            sage: C._initialize_fat_horizontal(p, 3)
            sage: C._reduce_vector_horizontal_BSGS((0, 0, 0, 0), 2*p  - 1, p)
            (0, 0, 0, 0)
            sage: C._reduce_vector_horizontal_BSGS((83283349998, 0, 0, 0), 2*p  - 1, p)
            (23734897071, 84632332850, 44254975407, 23684517017)
            sage: C._reduce_vector_horizontal_BSGS((98582524551, 3200841460, 6361495378, 98571346457), 2*p - 1, p)
            (96813533420, 61680190736, 123292559950, 96786566978)
        """
        if G == 0:
            return G
        if self._verbose > 2:
            print(
                "_reduce_vector_horizontal_BSGS(self, %s, %s, %s)"
                % (vector(self._Qq, G), e, s)
            )
        assert (e + 1) % self._p == 0
        (m0, m1), (M0, M1) = self._horizontal_matrix_reduction(s)
        vect = vector(self._Zq, G)
        # we do the first d reductions carefully
        D = 1
        for i in reversed(range(e - self._d + 1, e + 1)):
            Mi = M0 + i * M1
            Di = m0 + i * m1

            vect = Mi * vect
            D *= Di
        assert Di % self._p == 0
        iD = 1 / self._Zq0(D.lift() / self._p)
        vect = vector(self._Zq0, [iD * ZZ(elt.lift() / self._p) for elt in vect])
        # use BSGS

        iDH, MH = self._horizontal_fat_s[s][(e + 1) / self._p - 1]
        vect = iDH * (MH * vect.change_ring(self._Zq0))

        # last reduction
        i = e - self._p + 1
        Mi = M0 + i * M1
        Di = 1 / (m0 + i * m1)

        vect = Di * (Mi * vect.change_ring(self._Zq))

        if self._verbose > 2:
            print(
                "done _reduce_vector_horizontal_BSGS(self, %s, %s, %s)"
                % (vector(self._Qq, G), e, s)
            )
            print("return %s\n" % (vector(self._Qq, vect),))
        return vect

    def _initialize_fat_horizontal(self, s, L):
        """
        Initialise reduction matrices for horizontal reductions for blocks from `s` to `L`.

        TESTS::

            sage: p = 4999
            sage: x = PolynomialRing(GF(p), "x").gen()
            sage: C = CyclicCover(3, x^4 + 4*x^3 + 9*x^2 + 3*x + 1)
            sage: C._init_frob()
            sage: C._initialize_fat_horizontal(p, 3)
            sage: len(C._horizontal_fat_s[p])
            3
        """
        assert self._sqrtp
        if s not in self._horizontal_fat_s:
            N = self._N - 1  # padic precision of self._Zq0
            d = self._d
            L0 = min(L, N)
            targets = [0] * (2 * L0)
            for l in range(L0):
                targets[2 * l] = self._p * l
                targets[2 * l + 1] = self._p * (l + 1) - d - 1
            (m0, m1), (M0, M1) = self._horizontal_matrix_reduction(s)
            M0, M1 = (elt.change_ring(self._Zq0) for elt in [M0, M1])
            D0, D1 = (matrix(self._Zq0, [elt]) for elt in [m0, m1])
            MH = interval_products(M0, M1, targets)
            DH = [elt[0, 0] for elt in interval_products(D0, D1, targets)]
            if L > N:  # Vandermonde interpolation
                #  f^{(r)}(0) p^r / r! for r = 0, ..., N-1,
                MT = [None] * N
                DT = [0] * N
                for r in range(N):
                    MT[r] = matrix(self._Zq0, d, d)
                    for h in range(N):
                        v = self._vandermonde[r, h]
                        for i in range(d):
                            for j in range(d):
                                MT[r][i, j] += v * MH[h][i, j]
                        DT[r] += v * DH[h]
                for k in range(N, L):
                    M = matrix(self._Zq0, d, d)
                    D = 0
                    k1pow = self._Zq0(1)  # power of k + 1
                    for h in range(N):
                        for i in range(d):
                            for j in range(d):
                                M[i, j] += k1pow * MT[h][i, j]
                        D += k1pow * DT[h]
                        k1pow *= k + 1
                    MH.append(M)
                    DH.append(D)

            iDH = [1 / elt for elt in DH]
            self._horizontal_fat_s[s] = [(iDH[i], MH[i]) for i in range(L)]
        assert len(self._horizontal_fat_s[s]) >= L

    def _reduce_vector_horizontal_plain(self, G, e, s, k=1):
        r"""
        INPUT:

        - a vector -- `G \in W_{e, s}`

        OUTPUT:

        - a vector -- `H \in W_{e - k, s}` such that
            `G x^e y^{-s} dx \cong H x^{e - k} y^{-s} dx`

        TESTS::

            sage: p = 4999
            sage: x = PolynomialRing(GF(p), "x").gen()
            sage: C = CyclicCover(3, x^4 + 4*x^3 + 9*x^2 + 3*x + 1)
            sage: C._init_frob()
            sage: C._initialize_fat_horizontal(p, 3)
            sage: C._reduce_vector_horizontal_plain((83283349998, 0, 0, 0), 2*p  - 1, p, p)
            (23734897071, 110671913892, 91161207284, 49524178051)
            sage: C._reduce_vector_horizontal_plain((98582524551, 3200841460, 6361495378, 98571346457), 2*p - 1, p, p)
            (96813533420, 65678590896, 12037075498, 66773575777)
            sage: (C._reduce_vector_horizontal_plain((98582524551, 3200841460, 6361495378, 98571346457), 2*p - 1, p, p) - C._reduce_vector_horizontal_plain((98582524551, 3200841460, 6361495378, 98571346457), 2*p - 1, p, p)) % p^C._N0 == 0
            True
        """
        if self._verbose > 2:
            print(
                "_reduce_vector_horizontal_plain(self, G = %s, e = %s, s = %s, k = %s)"
                % (vector(self._Qq, G), e, s, k)
            )
        if G == 0:
            return G
        (m0, m1), (M0, M1) = self._horizontal_matrix_reduction(s)
        vect = vector(self._Zq, G)
        D = self._Zq(1)
        assert k <= self._p, "more than p reductions at a time should be avoided!"
        assert e - k + 1 >= 0
        for i in reversed(range(e - k + 1, e + 1)):
            Mi = M0 + i * M1
            Di = m0 + i * m1

            vect = Mi * vect
            D *= Di
            if self._plarge and Di % self._p == 0:
                assert (i + self._d) % self._p == 0
                vect = self._divide_vector(D, vect, self._Zq)
                D = self._Zq(1)

        vect = self._divide_vector(D, vect, self._Zq)

        if self._verbose > 2:
            print(
                "done _reduce_vector_horizontal_plain(self, %s, %s, %s, %s)"
                % (vector(self._Qq, G), e, s, k)
            )
            print("return %s\n" % (vector(self._Qq, vect),))
        return vect

    def _reduce_vector_vertical(self, G, s0, s, k=1):
        r"""
        Reduce the vector `G` representing an element of `W_{-1,rs + s0}` by `r k` steps.

        INPUT:

        - a vector -- `G \in W_{-1, r*s + s0}`

        OUTPUT:

        - a vector -- `H \in W_{-1, r*(s - k) + s0}` such that
          `G y^{-(r*s + s0)} dx \cong H y^{-(r*(s -k) + s0)} dx`

        TESTS::

            sage: p = 4999
            sage: x = PolynomialRing(GF(p),"x").gen()
            sage: C = CyclicCover(3, x^4 + 4*x^3 + 9*x^2 + 3*x + 1)
            sage: C._init_frob()
            sage: C._reduce_vector_vertical((96683034459, 33074345103, 43132216686), 1, p // 3, p // 3)
            (4364634303, 124117911400, 26239932330)
        """

        def _reduce_vector_vertical_plain(G, s0, s, k=1):
            r"""
            INPUT:

            - a vector -- `G \in W_{-1, r*s + s0}`

            OUTPUT:

            - a vector -- `H \in W_{-1, r*(s - k) + s0}` such that
              `G y^{-(r*s + s0)} dx \cong H y^{-(r*(s -k) + s0)} dx`
            """
            if self._verbose > 2:
                print(
                    "_reduce_vector_vertical(self, G = %s, s0 = %s, s = %s, k = %s)"
                    % (vector(self._Qq, G), s0, s, k)
                )

            (m0, m1), (M0, M1) = self._vertical_matrix_reduction(s0)
            vect = vector(self._Zq, G)
            D = 1
            assert k <= self._p, "more than p reductions at a time should be avoided!"
            assert s - k >= 0
            for i in reversed(range(s - k + 1, s + 1)):
                Mi = M0 + i * M1
                Di = m0 + i * m1
                vect = Mi * vect
                if self._plarge and Di % self._p != 0:
                    D *= Di
                else:
                    vect = self._divide_vector(Di, vect, self._Zq)

            vect = self._divide_vector(D, vect, self._Zq)

            if self._verbose > 2:
                print(
                    "done _reduce_vector_vertical(self,  %s, %s, %s)"
                    % (vector(self._Qq, G), s, k)
                )
                print("return %s\n" % (vector(self._Qq, vect),))

            return vect

        G = vector(self._Zq, G)

        if self._sqrtp:
            self._initialize_fat_vertical(s0, s)
            if k < self._p:
                assert s - k == self._epsilon
                MV = self._vertical_fat_s[s0][0]
            elif k == self._p:
                MV = self._vertical_fat_s[s0][s // self._p]

            return MV * G
        return _reduce_vector_vertical_plain(G, s0, s, k)

    def _initialize_fat_vertical(self, s0, max_upper_target):
        """
        Initialise reduction matrices for vertical reductions for blocks from `s0` to `s0 + max_upper_target`.

        TESTS::

            sage: p = 4999
            sage: x = PolynomialRing(GF(p), "x").gen()
            sage: C = CyclicCover(3, x^4 + 4*x^3 + 9*x^2 + 3*x + 1)
            sage: C._init_frob()
            sage: C._initialize_fat_vertical(1, p + p // 3)
            sage: len(C._vertical_fat_s[1])
            2
        """
        L = floor((max_upper_target - self._epsilon) / self._p) + 1
        if s0 not in self._vertical_fat_s:
            (m0, m1), (M0, M1) = self._vertical_matrix_reduction(s0)
            D0, D1 = (matrix(self._Zq, [y]) for y in [m0, m1])
            targets = [0] * (2 * L)
            for l in reversed(range(L)):
                targets[2 * l] = max_upper_target - self._p * (L - l)
                targets[2 * l + 1] = max_upper_target - self._p * (L - 1 - l)
            if targets[0] < 0:
                targets[0] = self._epsilon
            MV = interval_products(M0, M1, targets)
            DV = interval_products(D0, D1, targets)
            for l in range(L):
                D = DV[l][0, 0]
                if D % self._p == 0:
                    iD = 1 / self._Zq(D.lift() / self._p)
                    MV[l] = matrix(
                        self._Zq,
                        [
                            [iD * ZZ(elt.lift() / self._p) for elt in row]
                            for row in MV[l].rows()
                        ],
                    )
                else:
                    MV[l] *= 1 / D
            self._vertical_fat_s[s0] = MV

        assert len(self._vertical_fat_s[s0]) >= L

    def _frob(self, i, j, N0):
        r"""
        Compute `Frob(x^i dx/y^j) / dx` in terms of the cohomology basis,
        whose `x` and `y` exponents are constrained to be in a particular range.

        INPUT:

        - ``i``, ``j`` -- exponents of the basis differential
        - ``N0`` -- desired `p`-adic precision for the Frobenius expansion

        TESTS::

            sage: p = 4999
            sage: x = PolynomialRing(GF(p), "x").gen()
            sage: C = CyclicCover(3, x^4 + 4*x^3 + 9*x^2 + 3*x + 1)
            sage: C._init_frob()
            sage: C._frob(2, 0, 1)
            [3174 + O(4999)]
            [1844 + O(4999)]
            [4722 + O(4999)]
        """
        # a Matrix that represents the Frobenius expansion of
        # x^i dx/y^j modulo p^(N0 + 1)
        # the entry (l, s) corresponds to the coefficient associated to the monomial x ** (p * (i + 1 + l) -1) * y ** (p * (j + r*s))
        assert N0 <= self._N0
        frobij = self._frob_sparse(i, j, N0)
        # recall
        # Frob(x^i dx/y^j) / dx
        #       = p * x ** (p * (i+1) - 1) * y ** (j*p)
        #         * = \sum_{s = 0} ^{N0-1}
        #               \sum_{l = 0} ^(d*s)
        #                   D_{j, s} * Frobpow[s][l] * x ** (p ** l) y ** (r * p ** s)
        # H represents H(x) * y^(-p**s) s /dx
        # the entry (l, s) of frobij
        # corresponds to the monomial (p * (i + 1 + l) -1, p * -(j + r*s))
        d = self._d
        r = self._r
        p = self._p
        H = vector(self._Zq, d - 1)
        k = (p * j) // r
        s0 = (p * j) % r

        for s in reversed(range(N0)):
            if self._sqrtp:
                # (i + 1) <= d
                self._initialize_fat_horizontal(
                    p * j + p * r * s, d * s + (d - 2) + 1
                )  # d * (s + 1) )
            # G represents G(x) * x^(p ** l - 1) y^(-p(j + r*s)) /dx
            G = vector(self._Zq, d)
            for ell in reversed(range(1, d * s + (i + 1) + 1)):
                if ell >= (i + 1):
                    G[0] += frobij[ell - (i + 1), s]
                G = self._reduce_vector_horizontal(G, p * ell - 1, p * j + p * r * s, p)
            assert G[0] == 0
            H += vector(G.list()[1:])
            if s > 0:
                # y^(-p(j + r*s))  --- >  y^(-p(j + r*(s-1)))
                H = self._reduce_vector_vertical(H, s0, k + p * s, p)
        # H represents
        # H(x) y^-p*j = H(x) y^-(k*r + s0)
        # now we reduce the pole order to s0 + r*epsilon, where s0 = p *j % r
        while k > self._epsilon:
            steps = p if k - self._epsilon > p else k - self._epsilon
            H = self._reduce_vector_vertical(H, s0, k, steps)
            k -= steps
        assert k == self._epsilon
        H = [self._Qq(elt).add_bigoh(N0) for elt in H]
        return matrix(H).transpose()

    @cached_method
    def frobenius_matrix(self, N=None):
        """
        Compute `p`-adic Frobenius matrix to precision `p^N`.

        If `N` not supplied, a default value is selected, which is the minimum
        needed to recover the charpoly unambiguously.

        EXAMPLES::

            sage: p = 107
            sage: x = PolynomialRing(GF(p), "x").gen()
            sage: CyclicCover(2, x^5 + x).frobenius_matrix()
            [              O(107^2)      89*107 + O(107^2)               O(107^2)               O(107^2)]
            [     89*107 + O(107^2)               O(107^2)               O(107^2)               O(107^2)]
            [              O(107^2)               O(107^2)               O(107^2) 105 + 5*107 + O(107^2)]
            [              O(107^2)               O(107^2) 89 + 53*107 + O(107^2)               O(107^2)]
            sage: CyclicCover(2, 3*x^5 + x).frobenius_matrix()
            [              O(107^2)      14*107 + O(107^2)               O(107^2)               O(107^2)]
            [     69*107 + O(107^2)               O(107^2)               O(107^2)               O(107^2)]
            [              O(107^2)               O(107^2)               O(107^2) 61 + 58*107 + O(107^2)]
            [              O(107^2)               O(107^2) 69 + 53*107 + O(107^2)               O(107^2)]
            sage: CyclicCover(3, x^3 + x).frobenius_matrix()
            [          0           0      O(107)      O(107)]
            [          0           0 52 + O(107)      O(107)]
            [     O(107) 35 + O(107)           0           0]
            [44 + O(107)      O(107)           0           0]
            sage: CyclicCover(3, 3*x^3 + x).frobenius_matrix()
            [          0           0      O(107)      O(107)]
            [          0           0 79 + O(107)      O(107)]
            [     O(107) 42 + O(107)           0           0]
            [30 + O(107)      O(107)           0           0]
        """

        def _frobenius_matrix_p(N0):
            r"""
            Compute the matrix that represents the p-power Frobenius
            """
            assert self._init_frobQ

            m = matrix(self._Qq, (self._d - 1) * (self._r - 1))

            # Want to build m, "slice by slice" using the output of _frob
            for j in range(1, self._r):
                s0 = (j * self._p) % self._r
                for i in range(self._d - 1):
                    m[(s0 - 1) * (self._d - 1):s0 * (self._d - 1),
                      i + (j - 1) * (self._d - 1),
                      ] = self._frob(i, j + self._epsilon * self._r, N0)
            return m

        self._init_frob(N)
        FrobP = _frobenius_matrix_p(self._N0)
        assert N == self._N0 or N is None
        if self._n == 1:
            return FrobP
        current = FrobP
        total = FrobP
        for i in range(self._n - 1):
            current = matrix(
                [[entry.frobenius() for entry in row] for row in current]
            )
            total = total * current
        total = matrix([[elt.add_bigoh(self._N0) for elt in row] for row in total])
        return total

    @cached_method
    def frobenius_polynomial(self):
        r"""
        Return the characteristic polynomial of Frobenius.

        EXAMPLES:

        Hyperelliptic curves::

            sage: p = 11
            sage: x = PolynomialRing(GF(p), "x").gen()
            sage: f = x^7 + 4*x^2 + 10*x + 4
            sage: CyclicCover(2, f).frobenius_polynomial() == \
            ....: HyperellipticCurve(f).frobenius_polynomial()
            True
            sage: f = 2*x^5 + 4*x^3 + x^2 + 2*x + 1
            sage: CyclicCover(2, f).frobenius_polynomial() == \
            ....: HyperellipticCurve(f).frobenius_polynomial()
            True
            sage: f = 2*x^6 + 4*x^4 + x^3 + 2*x^2 + x
            sage: CyclicCover(2, f).frobenius_polynomial() == \
            ....: HyperellipticCurve(f).frobenius_polynomial()
            True
            sage: p = 1117
            sage: x = PolynomialRing(GF(p), "x").gen()
            sage: f = x^9 + 4*x^2 + 10*x + 4
            sage: P1 = CyclicCover(2, f).frobenius_polynomial()
            sage: P2 = HyperellipticCurve(f).frobenius_polynomial()
            sage: P1 == P2  # long time
            True
            sage: f = 2*x^5 + 4*x^3 + x^2 + 2*x + 1
            sage: CyclicCover(2, f).frobenius_polynomial() == \
            ....: HyperellipticCurve(f).frobenius_polynomial()
            True

        Superelliptic curves::

            sage: p = 11
            sage: x = PolynomialRing(GF(p), "x").gen()
            sage: CyclicCover(3, x^4 + 4*x^3 + 9*x^2 + 3*x + 1).frobenius_polynomial()
            x^6 + 21*x^4 + 231*x^2 + 1331
            sage: CyclicCover(4, x^3 + x + 1).frobenius_polynomial()
            x^6 + 2*x^5 + 11*x^4 + 121*x^2 + 242*x + 1331
            sage: p = 4999
            sage: x = PolynomialRing(GF(p),"x").gen()
            sage: CyclicCover(4, x^3  - 1).frobenius_polynomial() == \
            ....: CyclicCover(3, x^4  + 1).frobenius_polynomial()
            True
            sage: CyclicCover(3, x^4 + 4*x^3 + 9*x^2 + 3*x + 1).frobenius_polynomial()
            x^6 + 180*x^5 + 20988*x^4 + 1854349*x^3 + 104919012*x^2 + 4498200180*x + 124925014999
            sage: CyclicCover(4, x^5 + x + 1).frobenius_polynomial()
            x^12 - 64*x^11 + 5018*x^10 - 488640*x^9 + 28119583*x^8 - 641791616*x^7
             + 124245485932*x^6 - 3208316288384*x^5 + 702708407289583*x^4 - 61043359329111360*x^3
             + 3133741752599645018*x^2 - 199800079984001599936*x + 15606259372500374970001

            sage: h = PolynomialRing(GF(1129), 'x')([-1] + [0]*(5-1) + [1])
            sage: CyclicCover(11, h).frobenius_polynomial()  # long time
            x^40 + 7337188909826596*x^30 + 20187877911930897108199045855206*x^20
             + 24687045654725446027864774006541463602997309796*x^10
             + 11320844849639649951608809973589776933203136765026963553258401

            sage: h = PolynomialRing(GF(1009^2), 'x')([-1] + [0]*(5-1) + [1])
            sage: CyclicCover(3, h).frobenius_polynomial()  # long time
            x^8 + 532*x^7 - 2877542*x^6 - 242628176*x^5 + 4390163797795*x^4 - 247015136050256*x^3
             - 2982540407204025062*x^2 + 561382189105547134612*x + 1074309286591662654798721

        A non-monic example checking that :issue:`29015` is fixed::

            sage: a = 3
            sage: K.<s> = GF(83^3);
            sage: R.<x> = PolynomialRing(K)
            sage: h = s*x^4 + x*3 + 8
            sage: C = CyclicCover(a, h)
            sage: C.frobenius_polynomial()
            x^6 + 1563486*x^4 + 893980969482*x^2 + 186940255267540403

        Non-superelliptic curves::

            sage: p = 13
            sage: x = PolynomialRing(GF(p), "x").gen()
            sage: C = CyclicCover(4, x^4 + 1)
            sage: C.frobenius_polynomial()
            x^6 - 6*x^5 + 3*x^4 + 60*x^3 + 39*x^2 - 1014*x + 2197
            sage: R.<t> = PowerSeriesRing(Integers())
            sage: C.projective_closure().zeta_series(2, t)
            1 + 8*t + 102*t^2 + O(t^3)
            sage: C.frobenius_polynomial().reverse()(t)/((1-t)*(1-p*t)) + O(t^5)
            1 + 8*t + 102*t^2 + 1384*t^3 + 18089*t^4 + O(t^5)

            sage: x = PolynomialRing(GF(11), "x").gen()
            sage: CyclicCover(4, x^6 - 11*x^3 + 70*x^2 - x + 961).frobenius_polynomial()  # long time
            x^14 + 14*x^12 + 287*x^10 + 3025*x^8 + 33275*x^6 + 381997*x^4 + 2254714*x^2 + 19487171
            sage: x = PolynomialRing(GF(4999), "x").gen()
            sage: CyclicCover(4, x^6 - 11*x^3 + 70*x^2 - x + 961).frobenius_polynomial()  # long time
            x^14 - 4*x^13 - 2822*x^12 - 30032*x^11 + 37164411*x^10 - 152369520*x^9
             + 54217349361*x^8 - 1021791160888*x^7 + 271032529455639*x^6 - 3807714457169520*x^5
             + 4642764601604000589*x^4 - 18754988504199390032*x^3 - 8809934776794570547178*x^2
             - 62425037490001499880004*x + 78015690603129374475034999

            sage: p = 11
            sage: x = PolynomialRing(GF(p), "x").gen()
            sage: CyclicCover(3, 5*x^3 - 5*x + 13).frobenius_polynomial()
            x^2 + 11
            sage: CyclicCover(3, x^6 + x^4 - x^3 + 2*x^2 - x - 1).frobenius_polynomial()
            x^8 + 32*x^6 + 462*x^4 + 3872*x^2 + 14641
            sage: p = 4999
            sage: x = PolynomialRing(GF(p), "x").gen()
            sage: CyclicCover(3, 5*x^3 - 5*x + 13).frobenius_polynomial()
            x^2 - 47*x + 4999
            sage: CyclicCover(3, x^6 + x^4 - x^3 + 2*x^2 - x - 1).frobenius_polynomial()
            x^8 + 122*x^7 + 4594*x^6 - 639110*x^5 - 82959649*x^4 - 3194910890*x^3
             + 114804064594*x^2 + 15240851829878*x + 624500149980001

            sage: p = 11
            sage: x = PolynomialRing(GF(p), "x").gen()
            sage: CyclicCover(5, x^5 + x).frobenius_polynomial()  # long time
            x^12 + 4*x^11 + 22*x^10 + 108*x^9 + 503*x^8 + 1848*x^7 + 5588*x^6 + 20328*x^5
             + 60863*x^4 + 143748*x^3 + 322102*x^2 + 644204*x + 1771561
            sage: CyclicCover(5, 2*x^5 + x).frobenius_polynomial()  # long time
            x^12 - 9*x^11 + 42*x^10 - 108*x^9 - 47*x^8 + 1782*x^7 - 8327*x^6 + 19602*x^5
             - 5687*x^4 - 143748*x^3 + 614922*x^2 - 1449459*x + 1771561
            sage: p = 49999
            sage: x = PolynomialRing(GF(p), "x").gen()
            sage: CyclicCover(5, x^5 + x).frobenius_polynomial()  # long time
            x^12 + 299994*x^10 + 37498500015*x^8 + 2499850002999980*x^6
             + 93742500224997000015*x^4 + 1874812507499850001499994*x^2
             + 15623125093747500037499700001
            sage: CyclicCover(5, 2*x^5 + x).frobenius_polynomial() # long time
            x^12 + 299994*x^10 + 37498500015*x^8 + 2499850002999980*x^6
             + 93742500224997000015*x^4 + 1874812507499850001499994*x^2
             + 15623125093747500037499700001


        TESTS::

            sage: for _ in range(5): # long time
            ....:     fail = False
            ....:     p = random_prime(500, lbound=5)
            ....:     for i in range(1, 4):
            ....:         F = GF((p, i))
            ....:         Fx = PolynomialRing(F, 'x')
            ....:         b = F.random_element()
            ....:         while b == 0:
            ....:            b = F.random_element()
            ....:         E = EllipticCurve(F, [0, b])
            ....:         C1 = CyclicCover(3, Fx([-b, 0, 1]))
            ....:         C2 = CyclicCover(2, Fx([b, 0, 0, 1]))
            ....:         frob = [elt.frobenius_polynomial() for elt in [E, C1, C2]]
            ....:         if len(set(frob)) != 1:
            ....:             E
            ....:             C1
            ....:             C2
            ....:             frob
            ....:             fail = True
            ....:             break
            ....:     if fail:
            ....:       break
            ....: else:
            ....:     True
            True
        """
        self._init_frob()
        F = self.frobenius_matrix(self._N0)

        def _denominator():
            R = PolynomialRing(ZZ, "T")
            T = R.gen()
            denom = R(1)
            lc = self._f.list()[-1]
            if lc == 1:  # MONIC
                for i in range(2, self._delta + 1):
                    if self._delta % i == 0:
                        phi = euler_phi(i)
                        G = IntegerModRing(i)
                        ki = G(self._q).multiplicative_order()
                        denom = denom * (T ** ki - 1) ** (phi // ki)
                return denom
            # Non-monic
            x = PolynomialRing(self._Fq, "x").gen()
            f = x ** self._delta - lc
            L = f.splitting_field("a")
            roots = [r for r, _ in f.change_ring(L).roots()]
            roots_dict = {r: i for i, r in enumerate(roots)}
            rootsfrob = [L.frobenius_endomorphism(self._Fq.degree())(r) for r in roots]
            m = zero_matrix(len(roots))
            for i, r in enumerate(roots):
                m[i, roots_dict[rootsfrob[i]]] = 1
            return R(R(m.characteristic_polynomial()) // (T - 1))

        denom = _denominator()
        R = PolynomialRing(ZZ, "x")

        if self._nodenominators:
            min_val = 0
        else:
            # are there any denominators in F?
            min_val = min(self._Qq(elt).valuation() for row in F.rows() for elt in row)

        if min_val >= 0:
            prec = _N0_nodenominators(self._p, self._genus, self._n)
            charpoly_prec = [prec + i for i in reversed(range(1, self._genus + 1))] + [
                prec
            ] * (self._genus + 1)
            cp = charpoly_frobenius(F, charpoly_prec, self._p, 1, self._n, denom.list())
            return R(cp)
        cp = F.charpoly().reverse()
        denom = denom.reverse()
        PS = PowerSeriesRing(self._Zp, "T")
        cp = PS(cp) / PS(denom)
        cp = cp.padded_list(self._genus + 1)
        cpZZ = [None for _ in range(2 * self._genus + 1)]
        cpZZ[0] = 1
        cpZZ[-1] = self._p ** self._genus
        for i in range(1, self._genus + 1):
            cmod = cp[i]
            bound = binomial(2 * self._genus, i) * self._p ** (i * self._n * 0.5)
            localmod = self._p ** (ceil(log(bound, self._p)))
            c = cmod.lift() % localmod
            if c > bound:
                c = -(-cmod.lift() % localmod)
            cpZZ[i] = c
            if i != self._genus + 1:
                cpZZ[2 * self._genus - i] = c * self._p ** (self._genus - i)
        cpZZ.reverse()
        return R(cpZZ)
